from copy import deepcopy
import re
from typing import Final

from tomlkit import comment, document, nl, string, table, ws
from tomlkit.items import AoT, Array, InlineTable, Table, Trivia
from tomlkit.toml_document import TOMLDocument

from .pkg_manifest import (
    BinaryDeclType,
    BinaryFileDeclType,
    BlobDeclType,
    DistfileDeclType,
    EmulatorDeclType,
    EmulatorProgramDeclType,
    FetchRestrictionDeclType,
    PackageManifest,
    PackageMetadataDeclType,
    ProvisionableDeclType,
    ServiceLevelDeclType,
    SourceDeclType,
    ToolchainComponentDeclType,
    ToolchainDeclType,
    VendorDeclType,
)
from ..utils.toml import (
    extract_footer_comments,
    extract_header_comments,
    inline_table_with_spaces,
    sorted_table,
    str_array,
)

RE_INDENT_FIX: Final = re.compile(r"(?m)^    ([\"'{\[])")


# XXX: To workaround https://github.com/python-poetry/tomlkit/issues/290,
# post-process the output to have all leading 4-space indentation before
# strings, lists or tables replaced by 2-space ones.
def _fix_indent(s: str) -> str:
    return RE_INDENT_FIX.sub(r"  \1", s)


def dumps_canonical_package_manifest_toml(
    pm: PackageManifest,
) -> str:
    return _fix_indent(_dump_canonical_package_manifest_toml(pm).as_string())


def _dump_canonical_package_manifest_toml(
    pm: PackageManifest,
) -> TOMLDocument:
    x = pm.to_raw()
    doc = pm.raw_doc

    y = document()

    if doc is not None:
        if header_comments := extract_header_comments(doc):
            last_is_ws = False
            for c in header_comments:
                if c.startswith("#"):
                    last_is_ws = False
                    y.add(comment(c[1:].strip()))
                else:
                    last_is_ws = True
                    y.add(ws(c))

            if not last_is_ws:
                y.add(nl())

    y.add("format", string(x["format"]))

    dump_metadata_decl_into(y, x["metadata"])
    dump_distfile_decls_into(y, x["distfiles"])
    maybe_dump_binary_decls_into(y, x.get("binary"))
    maybe_dump_blob_decl_into(y, x.get("blob"))
    maybe_dump_emulator_decl_into(y, x.get("emulator"))
    maybe_dump_provisionable_decl_into(y, x.get("provisionable"))
    maybe_dump_source_decl_into(y, x.get("source"))
    maybe_dump_toolchain_decl_into(y, x.get("toolchain"))

    if doc is not None:
        if footer_comments := extract_footer_comments(doc):
            if footer_comments[0].startswith("#"):
                y.add(nl())

            for c in footer_comments:
                if c.startswith("#"):
                    y.add(comment(c[1:].strip()))
                else:
                    y.add(ws(c))

    return y


def dump_service_level_entry(x: ServiceLevelDeclType) -> Table:
    y = table()
    y.add("level", x["level"])
    if msgid := x.get("msgid"):
        y.add("msgid", string(msgid))
    if params := x.get("params"):
        y.add("params", sorted_table(params))
    return y


def dump_service_level_decls(x: list[ServiceLevelDeclType]) -> AoT:
    return AoT([dump_service_level_entry(i) for i in x])


def dump_metadata_decl(x: PackageMetadataDeclType) -> Table:
    y = table()
    y.add("desc", string(x["desc"]))
    y.add("vendor", dump_vendor_decl(x["vendor"]))
    if "slug" in x:
        y.add("slug", string(x["slug"]))
    if uv := x.get("upstream_version"):
        y.add("upstream_version", string(uv))
    if sl := x.get("service_level"):
        y.add(nl())
        y.add("service_level", dump_service_level_decls(sl))
    return y


def dump_metadata_decl_into(doc: TOMLDocument, x: PackageMetadataDeclType) -> None:
    doc.add(nl())
    doc.add("metadata", dump_metadata_decl(x))


def dump_vendor_decl(x: VendorDeclType) -> InlineTable:
    y = inline_table_with_spaces()
    with y:
        y.add("name", string(x["name"]))
        y.add("eula", string(x["eula"] if x["eula"] is not None else ""))
    return y


def dump_distfile_decls(x: list[DistfileDeclType]) -> AoT:
    return AoT([dump_distfile_entry(i) for i in x])


def dump_distfile_decls_into(doc: TOMLDocument, x: list[DistfileDeclType]) -> None:
    doc.add(nl())
    doc.add("distfiles", dump_distfile_decls(x))


def dump_distfile_entry(x: DistfileDeclType) -> Table:
    y = table()
    y.add("name", x["name"])
    if v := x.get("unpack"):
        y.add("unpack", string(v))
    y.add("size", x["size"])
    s = x.get("strip_components")
    if s is not None and s != 1:
        y.add("strip_components", s)
    if p := x.get("prefixes_to_unpack"):
        y.add("prefixes_to_unpack", str_array(p, multiline=len(p) > 1))
    if "urls" in x:
        # XXX: https://github.com/python-poetry/tomlkit/issues/290 prevents us
        # from using 2-space indentation for the array items for now.
        y.add("urls", str_array([str(i) for i in x["urls"]], multiline=True))
    if r := x.get("restrict"):
        # If `restrict` is a string, convert it to a list, fixing a common
        # oversight in package manifests.
        if isinstance(r, str):
            r = [r]
        y.add("restrict", [str(i) for i in r])
    if f := x.get("fetch_restriction"):
        y.add("fetch_restriction", dump_fetch_restriction(f))
    y.add("checksums", sorted_table(x["checksums"]))
    return y


def dump_fetch_restriction(x: FetchRestrictionDeclType) -> Table:
    y = table()
    y.add("msgid", x["msgid"])
    if "params" in x:
        y.add("params", sorted_table(x["params"]))
    return y


def dump_blob_decl(x: BlobDeclType) -> Table:
    y = table()
    y.add("distfiles", str_array(x["distfiles"], multiline=True))
    return y


def maybe_dump_blob_decl_into(doc: TOMLDocument, x: BlobDeclType | None) -> None:
    if x is None:
        return
    doc.add(nl())
    doc.add("blob", dump_blob_decl(x))


def dump_provisionable_decl(x: ProvisionableDeclType) -> Table:
    y = table()
    y.add("strategy", x["strategy"])
    y.add(
        "partition_map",
        sorted_table({str(k): v for k, v in x["partition_map"].items()}),
    )
    return y


def maybe_dump_provisionable_decl_into(
    doc: TOMLDocument,
    x: ProvisionableDeclType | None,
) -> None:
    if x is None:
        return
    doc.add(nl())
    doc.add("provisionable", dump_provisionable_decl(x))


def dump_binary_decl(x: BinaryFileDeclType, last: bool) -> Table:
    y = table()
    y.add("host", string(x["host"]))
    multiline_distfiles = len(x["distfiles"]) > 1
    y.add("distfiles", str_array(x["distfiles"], multiline=multiline_distfiles))
    if cmds := x.get("commands", {}):
        y.add("commands", sorted_table(cmds))
        if not last:
            y.add(nl())
    return y


def dump_binary_decls(x: list[BinaryFileDeclType]) -> AoT:
    return AoT([dump_binary_decl(elem, i == len(x) - 1) for i, elem in enumerate(x)])


def maybe_dump_binary_decls_into(doc: TOMLDocument, x: BinaryDeclType | None) -> None:
    if x is None:
        return
    doc.add("binary", dump_binary_decls(x))


def dump_emulator_program_decl(x: EmulatorProgramDeclType) -> Table:
    y = table()
    y.add("path", string(x["path"]))
    y.add("flavor", string(x["flavor"]))
    y.add("supported_arches", str_array(x["supported_arches"]))
    if "binfmt_misc" in x:
        y.add("binfmt_misc", string(x["binfmt_misc"]))
    return y


def dump_emulator_decl(x: EmulatorDeclType) -> Table:
    y = table()
    # Prefer `quirks` to `flavors`
    quirks = x.get("quirks")
    if quirks is None:
        quirks = x.get("flavors", [])
    y.add("quirks", str_array(quirks))
    y.add("programs", AoT([dump_emulator_program_decl(i) for i in x["programs"]]))
    return y


def maybe_dump_emulator_decl_into(
    doc: TOMLDocument, x: EmulatorDeclType | None
) -> None:
    if x is None:
        return
    doc.add(nl())
    doc.add("emulator", dump_emulator_decl(x))


def dump_source_decl(x: SourceDeclType) -> Table:
    y = table()
    multiline_distfiles = len(x["distfiles"]) > 1
    y.add("distfiles", str_array(x["distfiles"], multiline=multiline_distfiles))
    return y


def maybe_dump_source_decl_into(doc: TOMLDocument, x: SourceDeclType | None) -> None:
    if x is None:
        return
    doc.add(nl())
    doc.add("source", dump_source_decl(x))


def dump_toolchain_component_decl(x: ToolchainComponentDeclType) -> InlineTable:
    y = inline_table_with_spaces()
    with y:
        y.add("name", string(x["name"]))
        y.add("version", string(x["version"]))
    return y


def dump_toolchain_component_decls(x: list[ToolchainComponentDeclType]) -> Array:
    sorted_x = deepcopy(x)
    sorted_x.sort(key=lambda i: i["name"])
    return Array(
        [dump_toolchain_component_decl(i) for i in sorted_x],
        Trivia(),
        multiline=True,
    )


def dump_toolchain_decl(x: ToolchainDeclType) -> Table:
    y = table()
    y.add("target", string(x["target"]))
    # Prefer `quirks` to `flavors`
    quirks = x.get("quirks")
    if quirks is None:
        quirks = x.get("flavors", [])
    y.add("quirks", str_array(quirks))
    y.add("components", dump_toolchain_component_decls(x["components"]))
    if "included_sysroot" in x:
        y.add("included_sysroot", x["included_sysroot"])
    return y


def maybe_dump_toolchain_decl_into(
    doc: TOMLDocument,
    x: ToolchainDeclType | None,
) -> None:
    if x is None:
        return
    doc.add(nl())
    doc.add("toolchain", dump_toolchain_decl(x))



"""CLI integration tests for `ruyi admin build-package` (B8)."""

from __future__ import annotations

import pathlib

from tests.fixtures import IntegrationTestHarness


def _setup_project(tmp_path: pathlib.Path, recipe_body: str) -> pathlib.Path:
    proj = tmp_path / "recipes-proj"
    proj.mkdir()
    (proj / "ruyi-build-recipes.toml").write_text(
        'format = "v1"\n[project]\nname = "integ"\n'
    )
    (proj / "out").mkdir()
    recipe = proj / "pkg.star"
    recipe.write_text(recipe_body)
    return recipe


def test_admin_build_package_dry_run(
    tmp_path: pathlib.Path,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    recipe = _setup_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return ctx.subprocess(argv = ['/bin/true'])\n"
        "RUYI.build.schedule_build(build_it)\n",
    )

    result = ruyi_cli_runner("admin", "build-package", "--dry-run", str(recipe))
    assert result.exit_code == 0, result.stderr
    assert "build_it" in result.stdout


def test_admin_build_package_executes(
    tmp_path: pathlib.Path,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    recipe = _setup_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return ctx.subprocess(argv = ['/bin/true'])\n"
        "RUYI.build.schedule_build(build_it)\n",
    )
    result = ruyi_cli_runner("admin", "build-package", str(recipe))
    assert result.exit_code == 0, result.stderr


def test_admin_build_package_var_flag(
    tmp_path: pathlib.Path,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    recipe = _setup_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return ctx.subprocess(argv = ['/bin/echo', ctx.var('arch')])\n"
        "RUYI.build.schedule_build(build_it)\n",
    )
    result = ruyi_cli_runner(
        "admin",
        "build-package",
        "--dry-run",
        "-v",
        "arch=riscv64",
        str(recipe),
    )
    assert result.exit_code == 0, result.stderr
    assert "riscv64" in result.stdout


def test_admin_build_package_invalid_var(
    tmp_path: pathlib.Path,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    recipe = _setup_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return ctx.subprocess(argv = ['/bin/true'])\n"
        "RUYI.build.schedule_build(build_it)\n",
    )
    result = ruyi_cli_runner(
        "admin",
        "build-package",
        "--dry-run",
        "-v",
        "no_equals_here",
        str(recipe),
    )
    assert result.exit_code == 1


def test_admin_build_package_build_failure_exits_nonzero(
    tmp_path: pathlib.Path,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    recipe = _setup_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return ctx.subprocess(argv = ['/bin/false'])\n"
        "RUYI.build.schedule_build(build_it)\n",
    )
    result = ruyi_cli_runner("admin", "build-package", str(recipe))
    assert result.exit_code != 0



import argparse
from typing import TYPE_CHECKING

from ..cli.cmd import RootCommand
from ..i18n import _

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


class NewsCommand(
    RootCommand,
    cmd="news",
    has_subcommands=True,
    is_subcommand_required=False,
    has_main=True,
    help=_("List and read news items from configured repository"),
):
    _my_parser: "ArgumentParser | None" = None

    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        cls._my_parser = p

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .news import maybe_notify_unread_news

        assert cls._my_parser is not None
        cls._my_parser.print_help()
        maybe_notify_unread_news(cfg, True)

        return 0


class NewsListCommand(
    NewsCommand,
    cmd="list",
    help=_("List news items"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument(
            "--new",
            action="store_true",
            help=_("List unread news items only"),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .news import do_news_list

        only_unread: bool = args.new
        return do_news_list(
            cfg,
            only_unread,
        )


class NewsReadCommand(
    NewsCommand,
    cmd="read",
    help=_("Read news items"),
    description=_(
        "Outputs news item(s) to the console and mark as already read. Defaults to reading all unread items if no item is specified."
    ),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument(
            "--quiet",
            "-q",
            action="store_true",
            help=_("Do not output anything and only mark as read"),
        )
        p.add_argument(
            "item",
            type=str,
            nargs="*",
            help=_("Ordinal or ID of the news item(s) to read"),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .news import do_news_read

        quiet: bool = args.quiet
        items_strs: list[str] = args.item

        return do_news_read(
            cfg,
            quiet,
            items_strs,
        )



import shutil
import sys
from typing import Final, Iterable, NoReturn

from ..cli.user_input import pause_before_continuing
from ..i18n import _
from ..log import RuyiLogger, humanize_list


def has_cmd_in_path(cmd: str) -> bool:
    return shutil.which(cmd) is not None


_CMDS: Final = (
    "bzip2",
    "gunzip",
    "lz4",
    "tar",
    "xz",
    "zstd",
    "unzip",
    # commands used by the device provisioner
    "sudo",
    "dd",
    "fastboot",
)

_CMD_PRESENCE_MAP: Final[dict[str, bool]] = {}


def init_cmd_presence_map() -> None:
    _CMD_PRESENCE_MAP.clear()
    for cmd in _CMDS:
        _CMD_PRESENCE_MAP[cmd] = has_cmd_in_path(cmd)


def ensure_cmds(
    logger: RuyiLogger,
    cmds: Iterable[str],
    interactive_retry: bool = True,
) -> None | NoReturn:
    # only allow interactive retry if stdin is a TTY
    interactive_retry = interactive_retry and sys.stdin.isatty()

    while True:
        if not _CMD_PRESENCE_MAP or interactive_retry:
            init_cmd_presence_map()

        # in case any command's availability is not cached in advance
        for cmd in cmds:
            if cmd not in _CMD_PRESENCE_MAP:
                _CMD_PRESENCE_MAP[cmd] = has_cmd_in_path(cmd)

        absent_cmds = sorted(
            cmd for cmd in cmds if not _CMD_PRESENCE_MAP.get(cmd, False)
        )
        if not absent_cmds:
            return None

        cmds_str = humanize_list(absent_cmds, item_color="yellow")
        prompt = _(
            "The command(s) {cmds} cannot be found in PATH, which [yellow]ruyi[/] requires"
        ).format(cmds=cmds_str)
        if not interactive_retry:
            logger.F(prompt)
            logger.I(_("please install and retry"))
            sys.exit(1)

        logger.W(prompt)
        logger.I(
            _(
                "please install them and press [green]Enter[/] to retry, or [green]Ctrl+C[/] to exit"
            )
        )
        try:
            pause_before_continuing(logger)
        except EOFError:
            logger.I(_("exiting due to EOF"))
            sys.exit(1)
        except KeyboardInterrupt:
            logger.I(_("exiting due to keyboard interrupt"))
            sys.exit(1)



# Re-implementation of necessary XDG Base Directory Specification semantics
# without pyxdg, which is under LGPL and not updated for the latest spec
# revision (0.6 vs 0.8 released in 2021).

import os
import pathlib
from typing import Iterable, NamedTuple


class XDGPathEntry(NamedTuple):
    path: pathlib.Path
    is_global: bool


def _paths_from_env(env: str, default: str) -> Iterable[pathlib.Path]:
    v = os.environ.get(env, default)
    for p in v.split(":"):
        yield pathlib.Path(p)


class XDGBaseDir:
    def __init__(self, app_name: str) -> None:
        self.app_name = app_name

    @property
    def cache_home(self) -> pathlib.Path:
        v = os.environ.get("XDG_CACHE_HOME", "")
        return pathlib.Path(v) if v else pathlib.Path.home() / ".cache"

    @property
    def config_home(self) -> pathlib.Path:
        v = os.environ.get("XDG_CONFIG_HOME", "")
        return pathlib.Path(v) if v else pathlib.Path.home() / ".config"

    @property
    def data_home(self) -> pathlib.Path:
        v = os.environ.get("XDG_DATA_HOME", "")
        return pathlib.Path(v) if v else pathlib.Path.home() / ".local" / "share"

    @property
    def state_home(self) -> pathlib.Path:
        v = os.environ.get("XDG_STATE_HOME", "")
        return pathlib.Path(v) if v else pathlib.Path.home() / ".local" / "state"

    @property
    def config_dirs(self) -> Iterable[XDGPathEntry]:
        # from highest precedence to lowest
        for p in _paths_from_env("XDG_CONFIG_DIRS", "/etc/xdg"):
            yield XDGPathEntry(p, True)

    @property
    def data_dirs(self) -> Iterable[XDGPathEntry]:
        # from highest precedence to lowest
        for p in _paths_from_env("XDG_DATA_DIRS", "/usr/local/share/:/usr/share/"):
            yield XDGPathEntry(p, True)

    # derived info

    @property
    def app_cache(self) -> pathlib.Path:
        return self.cache_home / self.app_name

    @property
    def app_config(self) -> pathlib.Path:
        return self.config_home / self.app_name

    @property
    def app_data(self) -> pathlib.Path:
        return self.data_home / self.app_name

    @property
    def app_state(self) -> pathlib.Path:
        return self.state_home / self.app_name

    @property
    def app_config_dirs(self) -> Iterable[XDGPathEntry]:
        # from highest precedence to lowest
        yield XDGPathEntry(self.app_config, False)
        for e in self.config_dirs:
            yield XDGPathEntry(e.path / self.app_name, e.is_global)

    @property
    def app_data_dirs(self) -> Iterable[XDGPathEntry]:
        # from highest precedence to lowest
        yield XDGPathEntry(self.app_data, False)
        for e in self.data_dirs:
            yield XDGPathEntry(e.path / self.app_name, e.is_global)



from typing import Mapping


def is_running_in_ci(os_environ: Mapping[str, str]) -> bool:
    """Simplified and quick CI check meant for basic judgement."""
    if os_environ.get("CI", "") == "true":
        return True
    elif os_environ.get("TF_BUILD", "") == "True":
        return True
    return False


def probe_for_ci(os_environ: Mapping[str, str]) -> str | None:
    # https://www.appveyor.com/docs/environment-variables/
    if os_environ.get("APPVEYOR", "").lower() == "true":
        return "appveyor"
    # https://learn.microsoft.com/en-us/azure/devops/pipelines/build/variables?view=azure-devops&tabs=yaml#system-variables-devops-services
    elif os_environ.get("TF_BUILD", "") == "True":
        return "azure"
    # https://circleci.com/docs/variables/#built-in-environment-variables
    elif os_environ.get("CIRCLECI", "") == "true":
        return "circleci"
    # https://cirrus-ci.org/guide/writing-tasks/#environment-variables
    elif os_environ.get("CIRRUS_CI", "") == "true":
        return "cirrus"
    # https://gitea.com/gitea/act_runner/pulls/113
    # this should be checked before GHA because upstream maintains compatibility
    # with GHA by also providing GHA-style preset variables
    # TODO: also detect Forgejo
    elif os_environ.get("GITEA_ACTIONS", "") == "true":
        return "gitea"
    # https://gitee.com/help/articles/4358#article-header8
    elif "GITEE_PIPELINE_NAME" in os_environ:
        return "gitee"
    # https://docs.github.com/en/actions/writing-workflows/choosing-what-your-workflow-does/store-information-in-variables#default-environment-variables
    elif os_environ.get("GITHUB_ACTIONS", "") == "true":
        return "github"
    # https://docs.gitlab.com/ee/ci/variables/predefined_variables.html#predefined-variables
    elif os_environ.get("GITLAB_CI", "") == "true":
        return "gitlab"
    # https://www.jenkins.io/doc/book/pipeline/jenkinsfile/#using-environment-variables
    # may have false-negatives but likely no false-positives
    elif "JENKINS_URL" in os_environ:
        return "jenkins"
    # https://gitee.com/openeuler/mugen
    # seems nothing except $OET_PATH is guaranteed
    elif "OET_PATH" in os_environ:
        return "mugen"
    # there seems to be no designated marker for openQA, test a couple of
    # hopefully ubiquitous variables to avoid going through the entire key set
    elif "OPENQA_CONFIG" in os_environ or "OPENQA_URL" in os_environ:
        return "openqa"
    # https://docs.travis-ci.com/user/environment-variables/#default-environment-variables
    elif os_environ.get("TRAVIS", "") == "true":
        return "travis"
    # https://docs.koderover.com/zadig/Zadig%20v3.1/project/build/
    # https://github.com/koderover/zadig/blob/v3.1.0/pkg/microservice/jobexecutor/core/service/job.go#L117
    elif os_environ.get("ZADIG", "") == "true":
        return "zadig"
    elif os_environ.get("CI", "") == "true":
        return "unidentified"

    return None



#!/usr/bin/env python3

import argparse
import os
import shutil
import subprocess
import sys
import time

import semver
import tomlkit

OPS_CHOICES = (
    "alpha",  # Make an alpha pre-release
    "beta",  # Make a beta pre-release
    "release",  # Make an official release
    "date",  # Bump the datestamp
    "major",  # Bump the major version
    "minor",  # Bump the minor version
    "patch",  # Bump the patch version
    "commit",  # Actually make the changes and perform `git commit`
)


def main(argv: list[str]) -> int:
    a = argparse.ArgumentParser()
    a.add_argument(
        "ops",
        choices=OPS_CHOICES,
        metavar="OP",
        nargs="+",
        help="Bumping operation to make",
    )

    args = a.parse_args(argv[1:])
    ops: list[str] = args.ops

    # assume CWD is project root, which is guaranteed to be the case (see
    # end of file)

    # first read current version
    with open("pyproject.toml", "rb") as fp:
        pyproject = tomlkit.load(fp)
    curr_ver_str: str = pyproject["project"]["version"]  # type: ignore[assignment,index,unused-ignore]
    curr_ver = semver.Version.parse(curr_ver_str)
    prerelease = curr_ver.prerelease

    testing_kind: str | None = None
    datestamp: str | None = None
    if prerelease:
        testing_kind, datestamp = prerelease.split(".", 1)

    new_ver_components = curr_ver.to_dict()
    assert isinstance(new_ver_components["major"], int)
    assert isinstance(new_ver_components["minor"], int)
    assert isinstance(new_ver_components["patch"], int)
    commit = False
    for op in ops:
        match op:
            case "alpha" | "beta" | "date":
                datestamp = time.strftime("%Y%m%d")
                if op != "date":
                    testing_kind = op
                new_ver_components["prerelease"] = f"{testing_kind}.{datestamp}"
            case "release":
                new_ver_components["prerelease"] = None
            case "major":
                new_ver_components["major"] += 1
                new_ver_components["minor"] = 0
                new_ver_components["patch"] = 0
            case "minor":
                new_ver_components["minor"] += 1
                new_ver_components["patch"] = 0
            case "patch":
                new_ver_components["patch"] += 1
            case "commit":
                commit = True
                break
            case _:
                raise NotImplementedError(f"unhandled op '{op}'")

    new_ver = semver.Version(**new_ver_components)  # type: ignore[arg-type]
    new_ver_str = str(new_ver)
    if new_ver_str == curr_ver_str:
        # due to our adoption of CalVer in prerelease numbering, we have
        # to wait for another day if a release of the curr kind is
        # already made today
        print("error: version unchanged after bumping", file=sys.stderr)
        return 1

    if not commit:
        print(f"info: would bump {curr_ver_str} to {new_ver_str}")
        print('info: changes are not made; re-run with "commit" op')
        return 0

    print(f"info: bumping {curr_ver_str} to {new_ver_str}")

    git_path = shutil.which("git")
    if git_path is None:
        print("error: git not found in PATH", file=sys.stderr)
        return 1

    # TODO: ensure staging area is clean before touching files?
    _bump_pyproject_toml("pyproject.toml", pyproject, new_ver_str, False)
    _bump_pyproject_toml("contrib/poetry-1.x/pyproject.toml", None, new_ver_str, True)
    _bump_ruyi_version_py("ruyi/version.py", new_ver_str)

    touched_files = [
        "pyproject.toml",
        "contrib/poetry-1.x/pyproject.toml",
        "ruyi/version.py",
    ]
    subprocess.run([git_path, "add"] + touched_files, check=True)

    if sys.stdin.isatty():
        stdin_target = os.readlink(f"/proc/self/fd/{sys.stdin.fileno()}")
        print(f"info: setting GPG_TTY to {stdin_target} for git commit")
        os.environ["GPG_TTY"] = stdin_target

    commit_title = f"build: bump self version to {new_ver_str}"
    subprocess.run([git_path, "commit", "-s", "-m", commit_title], check=True)

    if sys.stdin.isatty():
        # display the commit we just made for manual inspection, in case
        # something gets unintentionally included, if the current session is
        # interactive
        sys.stdout.flush()
        os.execv(git_path, ["git", "show"])

    return 0


def _bump_pyproject_toml(
    file: str,
    obj: tomlkit.TOMLDocument | None,
    new_ver: str,
    poetry1: bool,
) -> None:
    if obj is None:
        with open(file, "rb") as fp:
            obj = tomlkit.load(fp)

    if poetry1:
        obj["tool"]["poetry"]["version"] = new_ver  # type: ignore[index,unused-ignore]
    else:
        obj["project"]["version"] = new_ver  # type: ignore[index,unused-ignore]

    with open(file, "wb") as fp:
        fp.write(tomlkit.dumps(obj).encode("utf-8"))


_RUYI_SEMVER_LINE_PREFIX = "RUYI_SEMVER: Final ="


def _bump_ruyi_version_py(file: str, new_ver: str) -> None:
    lines = []
    with open(file, "r") as fp:
        for line in fp:
            if not line.startswith(_RUYI_SEMVER_LINE_PREFIX):
                lines.append(line)
                continue
            # we want an escaped string that's to be quoted with double
            # quotes, but repr() by default gives us single quotes
            new_ver_escaped = repr(new_ver)[1:-1].replace('"', '\\"')
            lines.append(f'{_RUYI_SEMVER_LINE_PREFIX} "{new_ver_escaped}"\n')

    with open(file, "wb") as fp:
        fp.write("".join(lines).encode("utf-8"))


if __name__ == "__main__":
    # cd to project root
    os.chdir(os.path.join(os.path.dirname(__file__), ".."))
    sys.exit(main(sys.argv))



import abc
import os
from typing import Final, Mapping, Protocol, Sequence, runtime_checkable

import ruyi

ENV_DEBUG: Final = "RUYI_DEBUG"
ENV_EXPERIMENTAL: Final = "RUYI_EXPERIMENTAL"
ENV_FORCE_ALLOW_ROOT: Final = "RUYI_FORCE_ALLOW_ROOT"
ENV_TELEMETRY_OPTOUT_KEY: Final = "RUYI_TELEMETRY_OPTOUT"
ENV_VENV_ROOT_KEY: Final = "RUYI_VENV"

TRUTHY_ENV_VAR_VALUES: Final = {"1", "true", "x", "y", "yes"}


def is_env_var_truthy(env: Mapping[str, str], var: str) -> bool:
    if v := env.get(var):
        return v.lower() in TRUTHY_ENV_VAR_VALUES
    return False


@runtime_checkable
class ProvidesGlobalMode(Protocol):
    @property
    def argv0(self) -> str: ...

    @property
    def main_file(self) -> str: ...

    @property
    def self_exe(self) -> str: ...

    @property
    def is_debug(self) -> bool: ...

    @property
    def is_experimental(self) -> bool: ...

    @property
    def is_packaged(self) -> bool: ...

    @property
    def is_porcelain(self) -> bool: ...

    @property
    def is_telemetry_optout(self) -> bool: ...

    @property
    def is_cli_autocomplete(self) -> bool: ...

    @property
    def venv_root(self) -> str | None: ...


class GlobalModeProvider(metaclass=abc.ABCMeta):
    """
    Abstract base class for global mode providers.
    """

    @property
    @abc.abstractmethod
    def argv0(self) -> str:
        return ""

    @property
    @abc.abstractmethod
    def main_file(self) -> str:
        return ""

    @property
    @abc.abstractmethod
    def self_exe(self) -> str:
        return ""

    def record_self_exe(self, argv0: str, main_file: str, self_exe: str) -> None:
        pass

    @property
    @abc.abstractmethod
    def is_debug(self) -> bool:
        return False

    @property
    @abc.abstractmethod
    def is_experimental(self) -> bool:
        return False

    @property
    @abc.abstractmethod
    def is_packaged(self) -> bool:
        return False

    @property
    @abc.abstractmethod
    def is_porcelain(self) -> bool:
        return False

    @is_porcelain.setter
    @abc.abstractmethod
    def is_porcelain(self, v: bool) -> None:
        pass

    @property
    @abc.abstractmethod
    def is_telemetry_optout(self) -> bool:
        return False

    @property
    @abc.abstractmethod
    def is_cli_autocomplete(self) -> bool:
        return False

    @property
    @abc.abstractmethod
    def venv_root(self) -> str | None:
        return None


def _guess_porcelain_from_argv(argv: list[str]) -> bool:
    """
    Guess if the current invocation is a "porcelain" command based on the
    arguments passed, without requiring the ``argparse`` machinery to be
    completely initialized.
    """
    # If the first argument is `--porcelain`, we assume it's a porcelain command.
    # This is currently accurate as the porcelain flag is only possible at this
    # position right now.
    return len(argv) > 1 and argv[1] == "--porcelain"


def is_cli_completion_script_requested(argv: Sequence[str]) -> bool:
    for arg in argv:
        if arg.split("=", 1)[0] == "--output-completion-script":
            return True
    return False


def _probe_cli_autocomplete(env: Mapping[str, str]) -> bool:
    """
    Probe if the current invocation is an argcomplete subprocess based on the
    environment, without requiring the ``argparse`` machinery to be completely
    initialized.
    """
    return "_ARGCOMPLETE" in env


class EnvGlobalModeProvider(GlobalModeProvider):
    def __init__(
        self,
        env: Mapping[str, str] | None = None,
        argv: list[str] | None = None,
    ) -> None:
        if env is None:
            env = os.environ
        if argv is None:
            argv = []

        self._argv0 = ""
        self._main_file = ""
        self._self_exe = ""

        self._is_debug = is_env_var_truthy(env, ENV_DEBUG)
        self._is_experimental = is_env_var_truthy(env, ENV_EXPERIMENTAL)
        self._is_porcelain = _guess_porcelain_from_argv(argv)
        self._is_telemetry_optout = is_env_var_truthy(env, ENV_TELEMETRY_OPTOUT_KEY)

        # We have to lift this piece of implementation detail out of argcomplete,
        # as the argcomplete import is very costly in terms of startup time.
        self._is_cli_autocomplete = _probe_cli_autocomplete(env)

        self._venv_root = env.get(ENV_VENV_ROOT_KEY)

    @property
    def argv0(self) -> str:
        return self._argv0

    @property
    def main_file(self) -> str:
        return self._main_file

    @property
    def self_exe(self) -> str:
        return self._self_exe

    def record_self_exe(self, argv0: str, main_file: str, self_exe: str) -> None:
        self._argv0 = argv0
        self._main_file = main_file
        self._self_exe = self_exe

    @property
    def is_debug(self) -> bool:
        return self._is_debug

    @property
    def is_experimental(self) -> bool:
        return self._is_experimental

    @property
    def is_packaged(self) -> bool:
        return hasattr(ruyi, "__compiled__")

    @property
    def is_porcelain(self) -> bool:
        return self._is_porcelain

    @is_porcelain.setter
    def is_porcelain(self, v: bool) -> None:
        self._is_porcelain = v

    @property
    def is_telemetry_optout(self) -> bool:
        return self._is_telemetry_optout

    @property
    def is_cli_autocomplete(self) -> bool:
        return self._is_cli_autocomplete

    @property
    def venv_root(self) -> str | None:
        return self._venv_root



import argparse
from typing import TYPE_CHECKING

from ..cli.cmd import RootCommand
from ..i18n import _
from .list_filter import ListFilter, ListFilterAction

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


class ListCommand(
    RootCommand,
    cmd="list",
    has_subcommands=True,
    is_subcommand_required=False,
    has_main=True,
    help=_("List available packages in configured repository"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument(
            "--verbose",
            "-v",
            action="store_true",
            help=_("Also show details for every package"),
        )

        p.add_argument(
            "--all",
            action=ListFilterAction,
            nargs=0,
            dest="filters",
            help=_("Match and show all packages"),
        )

        # filter expressions
        p.add_argument(
            "--is-installed",
            action=ListFilterAction,
            nargs=1,
            dest="filters",
            help=_(
                "Match packages that are installed (y/true/1) or not installed (n/false/0)"
            ),
        )
        p.add_argument(
            "--category-contains",
            action=ListFilterAction,
            nargs=1,
            dest="filters",
            help=_(
                "Match packages from categories whose names contain the given string"
            ),
        )
        p.add_argument(
            "--category-is",
            action=ListFilterAction,
            nargs=1,
            dest="filters",
            help=_("Match packages from the given category"),
        )
        p.add_argument(
            "--name-contains",
            action=ListFilterAction,
            nargs=1,
            dest="filters",
            help=_("Match packages whose names contain the given string"),
        )

        if gc.is_experimental:
            p.add_argument(
                "--related-to-entity",
                action=ListFilterAction,
                nargs=1,
                dest="filters",
                help=_("Match packages related to the given entity"),
            )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .list import do_list

        verbose: bool = args.verbose
        filters: ListFilter = args.filters

        return do_list(
            cfg,
            filters=filters,
            verbose=verbose,
        )



import itertools
import os.path
from typing import TYPE_CHECKING, TypedDict, TypeGuard, cast

from ..cli import user_input
from ..config import GlobalConfig
from ..i18n import _
from ..log import RuyiLogger
from ..ruyipkg.atom import Atom, ExprAtom, SlugAtom
from ..ruyipkg.entity_provider import BaseEntity
from ..ruyipkg.host import get_native_host
from ..ruyipkg.install import do_install_atoms
from ..ruyipkg.pkg_manifest import (
    KNOWN_PARTITION_KINDS,
    PartitionKind,
    PartitionMapDecl,
)
from ..ruyipkg.composite_repo import CompositeRepo
from ..utils import mounts, prereqs

if TYPE_CHECKING:
    from ..ruyipkg.pkg_manifest import BoundPackageManifest


def get_variant_display_name(dev: BaseEntity, variant: BaseEntity) -> str:
    """Get the display name of a device variant."""
    if n := variant.display_name:
        return n
    return f"{dev.display_name} ({variant.data['variant_name']})"


def do_provision_interactive(config: GlobalConfig) -> int:
    log = config.logger

    # ensure ruyi repo is present, for good out-of-the-box experience
    mr = config.repo
    mr.ensure_git_repo()

    log.stdout(
        _(
            """
[bold green]RuyiSDK Device Provisioning Wizard[/]

This is a wizard intended to help you install a system on your device for your
development pleasure, all with ease.

You will be asked some questions that help RuyiSDK understand your device and
your intended configuration, then packages will be downloaded and flashed onto
the device's storage, that you should somehow make available on this host
system beforehand.

Note that, as Ruyi does not run as [yellow]root[/], but raw disk access is most likely
required to flash images, you should arrange to allow your user account [yellow]sudo[/]
access to necessary commands such as [yellow]dd[/]. Flashing will fail if the [yellow]sudo[/]
configuration does not allow so.
"""
        )
    )

    if not user_input.ask_for_yesno_confirmation(log, _("Continue?")):
        log.stdout(
            _("\nExiting. You can restart the wizard whenever prepared."),
            end="\n\n",
        )
        return 1

    device_entities = list(mr.entity_store.iter_entities("device"))
    device_entities.sort(key=lambda x: x.display_name or "")
    devices_by_id = {x.id: x for x in device_entities}

    dev_choices = {k: v.display_name or "" for k, v in devices_by_id.items()}
    dev_id = user_input.ask_for_kv_choice(
        log,
        _(
            "\nThe following devices are currently supported by the wizard. Please pick your device:"
        ),
        dev_choices,
    )
    dev = devices_by_id[dev_id]

    variants = list(
        mr.entity_store.traverse_related_entities(
            dev,
            entity_types=["device-variant"],
        )
    )
    variants.sort(key=lambda x: x.data.get("variant_name", ""))

    variant_choices = [get_variant_display_name(dev, i) for i in variants]
    variant_idx = user_input.ask_for_choice(
        log,
        _(
            "\nThe device has the following variants. Please choose the one corresponding to your hardware at hand:"
        ),
        variant_choices,
    )
    variant = variants[variant_idx]

    supported_combos = list(
        mr.entity_store.traverse_related_entities(
            variant,
            forward_refs=False,
            reverse_refs=True,
            entity_types=["image-combo"],
        )
    )
    supported_combos.sort(key=lambda x: x.display_name or "")
    combo_choices = [combo.display_name or "" for combo in supported_combos]
    combo_idx = user_input.ask_for_choice(
        log,
        _(
            "\nThe following system configurations are supported by the device variant you have chosen. Please pick the one you want to put on the device:"
        ),
        combo_choices,
    )
    combo = supported_combos[combo_idx]

    return do_provision_combo_interactive(config, mr, dev, variant, combo)


def maybe_render_postinst_msg(
    logger: RuyiLogger,
    mr: CompositeRepo,
    combo: BaseEntity,
    lang_code: str,
) -> bool:
    if postinst_msgid := combo.data.get("postinst_msgid"):
        # This field is named just "msgid" so no variables to render for
        # the retrieved text
        if msg := mr.messages.get_message_template(postinst_msgid, lang_code):
            logger.stdout(f"\n{msg}")
            return True
    return False


def do_provision_combo_interactive(
    config: GlobalConfig,
    mr: CompositeRepo,
    dev_decl: BaseEntity,
    variant_decl: BaseEntity,
    combo: BaseEntity,
) -> int:
    logger = config.logger
    devid = f"{dev_decl.id}@{variant_decl.id}"
    logger.D(f"provisioning device variant '{devid}'")

    # download packages
    pkg_atoms = combo.data.get("package_atoms", [])
    if not pkg_atoms:
        if maybe_render_postinst_msg(logger, mr, combo, config.lang_code):
            return 0

        logger.F(
            _(
                "malformed config: device variant '{devid}' asks for no packages but provides no messages either"
            ).format(
                devid=devid,
            )
        )
        return 1

    new_pkg_atoms = customize_package_versions(config, mr, pkg_atoms)
    if new_pkg_atoms is None:
        logger.stdout(
            _("\nExiting. You may restart the wizard at any time."), end="\n\n"
        )
        return 1
    else:
        pkg_atoms = new_pkg_atoms

    pkg_names_for_display = "\n".join(f" * [green]{i}[/]" for i in pkg_atoms)
    logger.stdout(
        _(
            "\nWe are about to download and install the following packages for your device:"
        ),
        end="\n\n",
    )
    logger.stdout(pkg_names_for_display, end="\n\n")

    if not user_input.ask_for_yesno_confirmation(logger, _("Proceed?")):
        logger.stdout(
            _("\nExiting. You may restart the wizard at any time."), end="\n\n"
        )
        return 1

    ret = do_install_atoms(
        config,
        mr,
        set(pkg_atoms),
        canonicalized_host=get_native_host(),
        fetch_only=False,
        reinstall=False,
    )
    if ret != 0:
        logger.F(_("failed to download and install packages"))
        logger.I(_("your device was not touched"))
        return 2

    strat_provider = ProvisionStrategyProvider(mr)
    strategies = [
        (pkg, get_pkg_provision_strategy(strat_provider, mr, pkg)) for pkg in pkg_atoms
    ]
    strategies.sort(key=lambda x: x[1].priority, reverse=True)

    # compose a partition map for each image pkg installed
    pkg_part_maps = {pkg: make_pkg_part_map(config, mr, pkg) for pkg in pkg_atoms}
    all_parts: list[PartitionKind] = []
    for pkg_part_map in pkg_part_maps.values():
        all_parts.extend(pkg_part_map.keys())

    # prompt user to give paths to target block device(s)
    requested_host_blkdevs: list[PartitionKind] = []
    for pkg, strat in strategies:
        requested_host_blkdevs.extend(strat.need_host_blkdevs(all_parts))

    host_blkdev_map: PartitionMapDecl = {}
    if requested_host_blkdevs:
        logger.stdout(
            _(
                """
For initializing this target device, you should plug into this host system the
device's storage (e.g. SD card or NVMe SSD), or a removable disk to be
reformatted as a live medium, and note down the corresponding device file
path(s), e.g. /dev/sdX, /dev/nvmeXnY for whole disks; /dev/sdXY, /dev/nvmeXnYpZ
for partitions. You may consult e.g. [yellow]sudo blkid[/] output for the
information you will need later.
"""
            )
        )
        for part in requested_host_blkdevs:
            part_desc = get_part_desc(part)

            while True:
                path = user_input.ask_for_file(
                    logger,
                    _("Please give the path for the {part_desc}:").format(
                        part_desc=part_desc,
                    ),
                )

                # Retrieve the latest mount info in case the user un-mounts
                # on seeing the prompt
                all_mounts = mounts.parse_mounts()
                blkdev_mounts = [m for m in all_mounts if m.source_is_blkdev]
                path_valid = True
                for m in blkdev_mounts:
                    if m.source_path.samefile(path):
                        logger.W(
                            _(
                                "path [cyan]'{path}'[/] is currently mounted at [yellow]'{target}'[/]"
                            ).format(
                                path=path,
                                target=m.target,
                            )
                        )
                        logger.I(
                            _(
                                "rejecting the path for safety; please double-check and retry"
                            )
                        )
                        path_valid = False
                        break
                if path_valid:
                    break

            host_blkdev_map[part] = path

    # final confirmation
    logger.stdout(
        _(
            """
We have collected enough information for the actual flashing. Now is the last
chance to re-check and confirm everything is fine.

We are about to:
"""
        )
    )

    pretend_steps = "\n".join(
        f" * {step_str}"
        for step_str in itertools.chain(
            *(
                strat[1].pretend(pkg_part_maps[strat[0]], host_blkdev_map)
                for strat in strategies
            )
        )
    )
    logger.stdout(pretend_steps, end="\n\n")

    if not user_input.ask_for_yesno_confirmation(logger, _("Proceed with flashing?")):
        logger.stdout(
            _(
                "\nExiting. The device is not touched and you may re-start the wizard at will."
            ),
            end="\n\n",
        )
        return 1

    # ensure commands
    all_needed_cmds = set(itertools.chain(*(strat.need_cmd for _, strat in strategies)))
    if all_needed_cmds:
        prereqs.ensure_cmds(logger, all_needed_cmds, interactive_retry=True)

        if "fastboot" in all_needed_cmds:
            # ask the user to ensure the device shows up
            # TODO: automate doing so
            logger.stdout(
                _("""
Some flashing steps require the use of fastboot, in which case you should
ensure the target device is showing up in [yellow]fastboot devices[/] output.
Please [bold red]confirm it yourself before continuing[/].
""")
            )
            if not user_input.ask_for_yesno_confirmation(
                logger,
                _("Is the device identified by fastboot now?"),
            ):
                logger.stdout(
                    _(
                        "\nAborting. The device is not touched. You may re-start the wizard after [yellow]fastboot[/] is fixed for the device."
                    ),
                    end="\n\n",
                )
                return 1

    # flash
    for pkg, strat in strategies:
        logger.D(f"flashing {pkg} with strategy {strat}")
        ret = strat.flash(pkg_part_maps[pkg], host_blkdev_map)
        if ret != 0:
            logger.F(_("flashing failed, check your device right now"))
            return ret

    # parting words
    logger.stdout(
        _("""
It seems the flashing has finished without errors.

[bold green]Happy hacking![/]
""")
    )

    maybe_render_postinst_msg(logger, mr, combo, config.lang_code)

    return 0


def get_part_desc(part: PartitionKind) -> str:
    match part:
        case "disk":
            return _("target's whole disk")
        case "live":
            return _("removable disk to use as live medium")
        case _:
            return _("target's '{part}' partition").format(part=part)


class PackageProvisionStrategyDecl(TypedDict):
    priority: int  # higher number means earlier
    need_host_blkdevs_fn: object  # Callable[[list[PartitionKind]], list[PartitionKind]]
    need_cmd: list[str]
    pretend_fn: object  # Callable[[PartitionMapDecl, PartitionMapDecl], list[str]]
    flash_fn: object  # Callable[[PartitionMapDecl, PartitionMapDecl], int]


def validate_list_str(x: object) -> TypeGuard[list[str]]:
    if not isinstance(x, list):
        return False
    x = cast(list[object], x)
    return all(isinstance(y, str) for y in x)


def validate_list_partition_kinds(x: object) -> TypeGuard[list[PartitionKind]]:
    if not isinstance(x, list):
        return False
    x = cast(list[object], x)
    for item in x:
        if not isinstance(item, str) or item not in KNOWN_PARTITION_KINDS:
            return False
    return True


class PackageProvisionStrategy:
    def __init__(
        self,
        decl: PackageProvisionStrategyDecl,
        mr: CompositeRepo,
    ) -> None:
        self._d = decl
        self._mr = mr

    @property
    def priority(self) -> int:
        return self._d["priority"]

    @property
    def need_cmd(self) -> list[str]:
        return self._d["need_cmd"]

    def need_host_blkdevs(self, x: list[PartitionKind]) -> list[PartitionKind]:
        result = self._mr.eval_plugin_fn(self._d["need_host_blkdevs_fn"], x)
        if not validate_list_partition_kinds(result):
            raise TypeError("need_host_blkdevs_fn must return list[PartitionKind]")
        return result

    def pretend(
        self,
        img_paths: PartitionMapDecl,
        blkdev_paths: PartitionMapDecl,
    ) -> list[str]:
        result = self._mr.eval_plugin_fn(self._d["pretend_fn"], img_paths, blkdev_paths)
        if not validate_list_str(result):
            raise TypeError("pretend_fn must return list[str]")
        return result

    def flash(
        self,
        img_paths: PartitionMapDecl,
        blkdev_paths: PartitionMapDecl,
    ) -> int:
        result = self._mr.eval_plugin_fn(self._d["flash_fn"], img_paths, blkdev_paths)
        if not isinstance(result, int):
            raise TypeError("flash_fn must return int")
        return result


class ProvisionStrategyProvider:
    def __init__(self, mr: CompositeRepo) -> None:
        self._mr = mr
        self._strats: dict[str, PackageProvisionStrategy] = {}

        # import the "standard library" of strategies
        self._import_strategy_plugin("std")

    def _import_strategy_plugin(self, plugin_pkg_name: str) -> None:
        plugin_id = f"ruyi-device-provision-strategy-{plugin_pkg_name}"
        provided_strats = self._mr.get_from_plugin(
            plugin_id,
            "PROVIDED_DEVICE_PROVISION_STRATEGIES_V1",
        )
        if not isinstance(provided_strats, dict):
            raise RuntimeError(
                f"malformed device provisioner strategy plugin '{plugin_id}'"
            )
        for name, decl in provided_strats.items():
            self._strats[name] = PackageProvisionStrategy(decl, self._mr)

    def __getitem__(self, name: str) -> PackageProvisionStrategy:
        try:
            return self._strats[name]
        except KeyError:
            # for now it's "ruyi-device-provision-strategy-STRATEGY-NAME"
            # we may have to revise before Ruyi v1.0 though
            self._import_strategy_plugin(name)
            return self._strats[name]


def get_pkg_provision_strategy(
    strat_provider: ProvisionStrategyProvider,
    mr: CompositeRepo,
    atom: str,
) -> PackageProvisionStrategy:
    a = Atom.parse(atom)
    pm = a.match_in_repo(mr, True)
    assert pm is not None

    pmd = pm.provisionable_metadata
    assert pmd is not None
    return strat_provider[pmd.strategy]


def make_pkg_part_map(
    config: GlobalConfig,
    mr: CompositeRepo,
    atom: str,
) -> PartitionMapDecl:
    a = Atom.parse(atom)
    pm = a.match_in_repo(mr, True)
    assert pm is not None
    pkg_root = config.global_blob_install_root(pm.name_for_installation)

    pmd = pm.provisionable_metadata
    assert pmd is not None
    return {p: os.path.join(pkg_root, f) for p, f in pmd.partition_map.items()}


def is_package_version_customization_possible(
    gc: GlobalConfig,
    mr: CompositeRepo,
    pkg_atoms: list[str],
) -> bool:
    """
    Check if package version customization is possible, which means there
    are at least one package atom specified that matches more than one versions.
    """

    for atom_str in pkg_atoms:
        # Get all available versions for this package
        a = Atom.parse(atom_str)
        try:
            if len(list(a.iter_in_repo(mr, gc.include_prereleases))) > 1:
                return True
        except KeyError:
            continue

    return False


def customize_package_versions(
    config: GlobalConfig,
    mr: CompositeRepo,
    pkg_atoms: list[str],
) -> list[str] | None:
    """
    Allow the user to customize the versions of packages to be installed.
    Returns a new list of package atoms with user-selected versions.
    """

    if not is_package_version_customization_possible(config, mr, pkg_atoms):
        return pkg_atoms

    logger = config.logger

    # Ask if the user wants to customize package versions
    logger.stdout(
        _(
            "By default, we'll install the latest version of each package, but in this case, other choices are possible."
        )
    )
    if not user_input.ask_for_yesno_confirmation(
        logger,
        _("Would you like to customize package versions?"),
    ):
        return pkg_atoms

    while True:  # Loop to allow restarting the selection process
        result: list[str] = []
        logger.stdout(_("\n[bold]Package Version Selection[/]"))

        for atom_str in pkg_atoms:
            # Parse the atom to get package name
            a = Atom.parse(atom_str)
            if isinstance(a, ExprAtom):
                # If it's already an expression with version constraints, show the constraints
                logger.stdout(
                    _(
                        "\nPackage [green]{atom}[/] already has version constraints."
                    ).format(
                        atom=atom_str,
                    )
                )
                if not user_input.ask_for_yesno_confirmation(
                    logger,
                    _("Would you like to change them?"),
                ):
                    result.append(atom_str)
                    continue
            elif isinstance(a, SlugAtom):
                # Slugs already fix the version, so we can't change them
                logger.W(
                    _(
                        "version cannot be overridden for slug atom [green]{atom}[/]"
                    ).format(
                        atom=atom_str,
                    )
                )
                result.append(atom_str)
                continue

            # Get all available versions for this package
            package_name = a.name
            category = a.category
            pkg_fullname = f"{category}/{package_name}" if category else package_name

            available_versions: "list[BoundPackageManifest]" = []
            try:
                available_versions = list(mr.iter_pkg_vers(package_name, category))
            except KeyError:
                logger.W(
                    _("could not find package [yellow]{pkg}[/] in repository").format(
                        pkg=pkg_fullname
                    )
                )
                result.append(atom_str)

            if not available_versions:
                logger.W(
                    _("no versions found for package [yellow]{pkg}[/]").format(
                        pkg=pkg_fullname
                    )
                )
                result.append(atom_str)
                continue

            if len(available_versions) == 1:
                # If there's only one version available, use it
                selected_version = available_versions[0]
                logger.stdout(
                    _(
                        "Only one version available for [green]{pkg}[/]: [blue]{ver}[/], using it."
                    ).format(
                        pkg=pkg_fullname,
                        ver=selected_version.ver,
                    )
                )
                result.append(atom_str)
                continue

            # Sort versions with newest first
            available_versions.sort(key=lambda pm: pm.semver, reverse=True)

            # Create a list of version choices for display
            version_choices = []
            for pm in available_versions:
                version_str = str(pm.semver)
                remarks = []

                if pm.is_prerelease:
                    remarks.append(_("prerelease"))
                if pm.service_level.has_known_issues:
                    remarks.append(_("has known issues"))
                if pm.upstream_version:
                    remarks.append(
                        _("upstream: {upstream_ver}").format(
                            upstream_ver=pm.upstream_version
                        )
                    )

                remark_str = f" ({', '.join(remarks)})" if remarks else ""
                version_choices.append(f"{version_str}{remark_str}")

            # Ask the user to select a version
            version_idx = user_input.ask_for_choice(
                logger,
                _("\nSelect a version for package [green]{pkg}[/]:").format(
                    pkg=pkg_fullname,
                ),
                version_choices,
            )

            selected_version = available_versions[version_idx]

            # Create the new atom string with the selected version
            if category:
                new_atom = f"{category}/{package_name}(=={selected_version.ver})"
            else:
                new_atom = f"{package_name}(=={selected_version.ver})"

            logger.stdout(_("Selected: [blue]{new_atom}[/]").format(new_atom=new_atom))
            result.append(new_atom)

        logger.stdout(_("\nPackage versions to be installed:"))
        for atom in result:
            logger.stdout(f" * [green]{atom}[/]")

        confirmation = user_input.ask_for_choice(
            logger,
            _("\nHow would you like to proceed?"),
            [
                _("Continue with these versions"),
                _("Restart version selection"),
                _("Abort device provisioning"),
            ],
        )

        if confirmation == 0:  # Continue with these versions
            return result
        elif confirmation == 1:  # Restart version selection
            logger.stdout(_("\nRestarting package version selection..."))
            continue
        else:  # Abort installation
            return None



#!/usr/bin/env python3

import sys
import time

# import these common stdlib modules beforehand, to help reduce clutter
# these must be modules that does not significantly affect the ruyi CLI's
# startup performance
STDLIBS_TO_PRELOAD = [
    "_struct",
    "argparse",
    "base64",
    "binascii",
    "bz2",
    "datetime",
    "functools",
    "gettext",
    "itertools",
    "lzma",
    "pathlib",
    "platform",
    "shutil",
    "struct",
    "typing",
    "os",
    "zlib",
]

if sys.version_info >= (3, 14):
    STDLIBS_TO_PRELOAD.append("annotationlib")

CURRENT_ALLOWLIST = {
    "ruyi",
    "ruyi.cli",
    "ruyi.cli.builtin_commands",
    "ruyi.cli.cmd",
    "ruyi.cli.completion",
    "ruyi.cli.config_cli",
    "ruyi.cli.self_cli",
    "ruyi.cli.version_cli",
    "ruyi.device",
    "ruyi.device.provision_cli",
    "ruyi.i18n",
    "ruyi.mux",
    "ruyi.mux.venv",
    "ruyi.mux.venv.venv_cli",
    "ruyi.pluginhost",
    "ruyi.pluginhost.plugin_cli",
    "ruyi.resource_bundle",
    "ruyi.resource_bundle.data",
    "ruyi.ruyipkg",
    "ruyi.ruyipkg.admin_cli",
    "ruyi.ruyipkg.cli_completion",  # part of the argparse machinery
    "ruyi.ruyipkg.entity_cli",
    "ruyi.ruyipkg.host",  # light-weight enough
    "ruyi.ruyipkg.install_cli",
    "ruyi.ruyipkg.list_cli",
    "ruyi.ruyipkg.list_filter",  # part of the argparse machinery
    "ruyi.ruyipkg.news_cli",
    "ruyi.ruyipkg.profile_cli",
    "ruyi.ruyipkg.repo_cli",
    "ruyi.ruyipkg.update_cli",
    "ruyi.telemetry",
    "ruyi.telemetry.telemetry_cli",
    "ruyi.utils",
    "ruyi.utils.global_mode",  # light-weight enough
}


def main() -> int:
    for lib in STDLIBS_TO_PRELOAD:
        __import__(lib)

    before = set(sys.modules.keys())
    a = time.monotonic_ns()

    from ruyi.cli import builtin_commands

    b = time.monotonic_ns()
    print(f"Import of built-in commands took {((b - a) / 1_000_000):.2f} ms.")
    del builtin_commands

    after = set(sys.modules.keys())
    modules_brought_in = after - before
    unwanted_modules = modules_brought_in - CURRENT_ALLOWLIST
    if not unwanted_modules:
        return 0

    print(
        """\
Some previously unneeded modules are now imported during built-in commands
initialization:
"""
    )
    for module in sorted(unwanted_modules):
        print(f"  - {module}")
    print(
        """
Please assess the impact on CLI startup performance before:

- allowing the module(s) by revising this script, or
- deferring the import(s) so they do not slow down CLI startup.
"""
    )
    return 1


if __name__ == "__main__":
    sys.exit(main())



import calendar
import datetime
import functools
import json
import pathlib
import time
from typing import Callable, TYPE_CHECKING, cast
import uuid

from ..i18n import _, d_
from ..log import RuyiLogger
from ..utils.node_info import NodeInfo, gather_node_info
from .scope import TelemetryScope
from .store import TelemetryStore

if TYPE_CHECKING:
    # for avoiding circular import
    from ..config import GlobalConfig

FALLBACK_PM_TELEMETRY_ENDPOINT = "https://api.ruyisdk.cn/telemetry/pm/"

TELEMETRY_CONSENT_AND_UPLOAD_DESC = d_(
    """
RuyiSDK collects minimal usage data in the form of just a version number of
the running [yellow]ruyi[/], to help us improve the product. With your consent,
RuyiSDK may also collect additional non-tracking usage data to be sent
periodically. The data will be recorded and processed by RuyiSDK team-managed
servers located in the Chinese mainland.

[green]By default, nothing leaves your machine[/], and you can also turn off usage data
collection completely. Only with your explicit permission can [yellow]ruyi[/] collect and
upload more usage data. You can change this setting at any time by running
[yellow]ruyi telemetry consent[/], [yellow]ruyi telemetry local[/], or [yellow]ruyi telemetry optout[/].

We'll also send a one-time report from this [yellow]ruyi[/] installation so the RuyiSDK
team can better understand adoption. If you choose to opt out, this will be the
only data to be ever uploaded, without any tracking ID being generated or kept.
Thank you for helping us build a better experience!
"""
)
TELEMETRY_CONSENT_AND_UPLOAD_PROMPT = d_(
    "Do you agree to have usage data periodically uploaded?"
)
TELEMETRY_OPTOUT_PROMPT = d_("\nDo you want to opt out of telemetry entirely?")
MALFORMED_TELEMETRY_STATE_MSG = d_(
    "malformed telemetry state: unable to determine upload weekday, nothing will be uploaded"
)


def next_utc_weekday(wday: int, now: float | None = None) -> int:
    t = time.gmtime(now)
    mday_delta = wday - t.tm_wday
    if mday_delta < 0:
        mday_delta += 7

    next_t = (
        t.tm_year,
        t.tm_mon,
        t.tm_mday + mday_delta,
        0,  # tm_hour
        0,  # tm_min
        0,  # tm_sec
        0,  # tm_wday
        0,  # tm_yday
        -1,  # tm_isdst
    )
    return calendar.timegm(next_t)


def set_telemetry_mode(
    gc: "GlobalConfig",
    mode: str,
    consent_time: datetime.datetime | None = None,
    show_cli_feedback: bool = True,
) -> None:
    """Set telemetry mode and consent time (if applicable) in the user preference."""

    from ..config.editor import ConfigEditor
    from ..config import schema

    logger = gc.logger

    if mode == "on":
        if consent_time is None:
            consent_time = datetime.datetime.now().astimezone()
    else:
        # clear any previously recorded consent time
        consent_time = None

    # First, persist the changes to user config
    with ConfigEditor.work_on_user_local_config(gc) as ed:
        ed.set_value((schema.SECTION_TELEMETRY, schema.KEY_TELEMETRY_MODE), mode)
        if consent_time is not None:
            ed.set_value(
                (schema.SECTION_TELEMETRY, schema.KEY_TELEMETRY_UPLOAD_CONSENT),
                consent_time,
            )
        else:
            ed.unset_value(
                (schema.SECTION_TELEMETRY, schema.KEY_TELEMETRY_UPLOAD_CONSENT)
            )

        ed.stage()

    # Then, apply the changes to the running instance's GlobalConfig
    # TelemetryProvider instance (if any) will pick them up automatically
    # because the properties are backed by GlobalConfig.
    gc.set_by_key(
        (schema.SECTION_TELEMETRY, schema.KEY_TELEMETRY_MODE),
        mode,
    )
    gc.set_by_key(
        (schema.SECTION_TELEMETRY, schema.KEY_TELEMETRY_UPLOAD_CONSENT),
        consent_time,
    )

    if not show_cli_feedback:
        return
    match mode:
        case "on":
            logger.I(_("telemetry data uploading is now enabled"))
            logger.I(
                _(
                    "you can opt out at any time by running [yellow]ruyi telemetry optout[/]"
                )
            )
        case "local":
            logger.I(_("telemetry mode is now set to local collection only"))
            logger.I(
                _(
                    "you can re-enable telemetry data uploading at any time by running [yellow]ruyi telemetry consent[/]"
                )
            )
            logger.I(
                _("or opt out at any time by running [yellow]ruyi telemetry optout[/]")
            )
        case "off":
            logger.I(_("telemetry data collection is now disabled"))
            logger.I(
                _(
                    "you can re-enable telemetry data uploads at any time by running [yellow]ruyi telemetry consent[/]"
                )
            )
        case _:
            raise ValueError(f"invalid telemetry mode: {mode}")


class TelemetryProvider:
    def __init__(self, gc: "GlobalConfig", minimal: bool) -> None:
        self.state_root = pathlib.Path(gc.telemetry_root)

        self._discard_events = False
        self._gc = gc
        self._is_first_run = False
        self._stores: dict[TelemetryScope, TelemetryStore] = {}
        self._upload_on_exit = False
        self.minimal = minimal

        # create the PM store
        self.init_store(TelemetryScope(None))
        # create per-repo stores for all active repos
        for entry in gc.repo_entries:
            if entry.active:
                self.init_store(TelemetryScope(entry.id))

    @property
    def logger(self) -> RuyiLogger:
        return self._gc.logger

    @property
    def local_mode(self) -> bool:
        return self._gc.telemetry_mode == "local"

    @property
    def upload_consent_time(self) -> datetime.datetime | None:
        if self.minimal or self.local_mode:
            return None
        return self._gc.telemetry_upload_consent_time

    def store(self, scope: TelemetryScope) -> TelemetryStore | None:
        return self._stores.get(scope)

    def init_store(self, scope: TelemetryScope) -> None:
        store_root = self.state_root
        api_url_fn: Callable[[], str | None] | None = None
        repo_name = scope.repo_name
        if repo_name is not None:
            repo_name_str: str = repo_name
            store_root = store_root / "repos" / repo_name_str

            def _f(rn: str = repo_name_str) -> str | None:
                # access the repo attribute lazily to speed up CLI startup
                return self._gc.repo.get_telemetry_api_url("repo", repo_id=rn)

            api_url_fn = _f
        else:
            # configure the PM telemetry endpoint
            api_url_fn = functools.partial(self._detect_pm_api_url, self._gc)

        store = TelemetryStore(
            self.logger,
            scope,
            store_root,
            api_url_factory=api_url_fn,
        )
        self._stores[scope] = store

    def _detect_pm_api_url(self, gc: "GlobalConfig") -> str | None:
        url = FALLBACK_PM_TELEMETRY_ENDPOINT
        cfg_src = "fallback"
        if gc.override_pm_telemetry_url is not None:
            cfg_src = "local config"
            url = gc.override_pm_telemetry_url
        else:
            if repo_provided_url := gc.repo.get_telemetry_api_url("pm"):
                cfg_src = "repo"
                url = repo_provided_url
        self.logger.D(
            f"configured PM telemetry endpoint via {cfg_src}: {url or '(n/a)'}"
        )
        return url

    @property
    def installation_file(self) -> pathlib.Path:
        return self.state_root / "installation.json"

    @property
    def minimal_installation_marker_file(self) -> pathlib.Path:
        return self.state_root / "minimal-installation-marker"

    def check_first_run_status(self) -> None:
        """Check if this is the first run of the application by checking if installation file exists.
        This must be done before init_installation() is potentially called.
        """
        self._is_first_run = (
            not self.installation_file.exists()
            and not self.minimal_installation_marker_file.exists()
        )

    @property
    def is_first_run(self) -> bool:
        """Check if this is the first run of the application."""
        return self._is_first_run

    def init_installation(self, force_reinit: bool) -> NodeInfo | None:
        if self.minimal:
            # be extra safe by not reading or writing installation data at all
            # in minimal mode
            self._init_minimal_installation_marker(force_reinit)
            return None

        installation_file = self.installation_file
        if installation_file.exists() and not force_reinit:
            return self._read_installation_data()

        # either this is a fresh installation or we're forcing a refresh
        installation_id = uuid.uuid4()
        self.logger.D(
            f"initializing telemetry data store, installation_id={installation_id.hex}"
        )
        self.state_root.mkdir(parents=True, exist_ok=True)

        # (over)write installation data
        installation_data = gather_node_info(installation_id)
        with open(installation_file, "wb") as fp:
            fp.write(json.dumps(installation_data).encode("utf-8"))
        return installation_data

    def _init_minimal_installation_marker(self, force_reinit: bool) -> None:
        if self.minimal_installation_marker_file.exists() and not force_reinit:
            return

        self.logger.D("initializing minimal installation marker file")
        self.state_root.mkdir(parents=True, exist_ok=True)

        # just touch the file
        self.minimal_installation_marker_file.touch()

    def _read_installation_data(self) -> NodeInfo | None:
        with open(self.installation_file, "rb") as fp:
            return cast(NodeInfo, json.load(fp))

    def _upload_weekday(self) -> int | None:
        if self.minimal:
            return None

        try:
            installation_data = self._read_installation_data()
        except FileNotFoundError:
            # init the node info if it's gone
            installation_data = self.init_installation(False)

        if installation_data is None:
            return None

        try:
            report_uuid_prefix = int(installation_data["report_uuid"][:8], 16)
        except ValueError:
            return None

        return report_uuid_prefix % 7  # 0 is Monday

    def _has_upload_consent(self, time_now: float | None = None) -> bool:
        if self.upload_consent_time is None:
            return False
        if time_now is None:
            time_now = time.time()
        return self.upload_consent_time.timestamp() <= time_now

    def _print_upload_schedule_notice(self, upload_wday: int, now: float) -> None:
        next_upload_day_ts = next_utc_weekday(upload_wday, now)
        next_upload_day = time.localtime(next_upload_day_ts)
        next_upload_day_end = time.localtime(next_upload_day_ts + 86400)
        next_upload_day_str = time.strftime("%Y-%m-%d %H:%M:%S %z", next_upload_day)
        next_upload_day_end_str = time.strftime(
            "%Y-%m-%d %H:%M:%S %z",
            next_upload_day_end,
        )

        if self._is_upload_day(now):
            for scope, store in self._stores.items():
                has_uploaded_today = self._has_uploaded_today(
                    store.last_upload_timestamp,
                    now,
                )
                if has_uploaded_today:
                    if last_upload_time := store.last_upload_timestamp:
                        last_upload_time_str = time.strftime(
                            "%Y-%m-%d %H:%M:%S %z", time.localtime(last_upload_time)
                        )
                        self.logger.I(
                            _(
                                "scope {scope}: usage information has already been uploaded today at {last_upload_time_str}"
                            ).format(
                                scope=scope,
                                last_upload_time_str=last_upload_time_str,
                            )
                        )
                    else:
                        self.logger.I(
                            _(
                                "scope {scope}: usage information has already been uploaded sometime today"
                            ).format(
                                scope=scope,
                            )
                        )
                else:
                    self.logger.I(
                        _(
                            "scope {scope}: the next upload will happen [bold green]today[/] if not already"
                        ).format(
                            scope=scope,
                        )
                    )
        else:
            self.logger.I(
                _("the next upload will happen anytime [yellow]ruyi[/] is executed:")
            )
            self.logger.I(
                _(
                    "  - between [bold green]{time_start}[/] and [bold green]{time_end}[/]"
                ).format(
                    time_start=next_upload_day_str,
                    time_end=next_upload_day_end_str,
                )
            )
            self.logger.I(_("  - or if the last upload is more than a week ago"))

    def print_telemetry_notice(self, for_cli_verbose_output: bool = False) -> None:
        if self.minimal:
            if for_cli_verbose_output:
                self.logger.I(
                    _(
                        "telemetry mode is [green]off[/]: nothing is collected or uploaded after the first run"
                    )
                )
            return

        now = time.time()
        upload_wday = self._upload_weekday()
        if upload_wday is None:
            if for_cli_verbose_output:
                self.logger.W(MALFORMED_TELEMETRY_STATE_MSG)
            else:
                self.logger.D(MALFORMED_TELEMETRY_STATE_MSG)
            return

        upload_wday_name = self._gc.babel_locale.days["format"]["wide"][upload_wday]

        if self.local_mode:
            if for_cli_verbose_output:
                self.logger.I(
                    _(
                        "telemetry mode is [green]local[/]: local usage collection only, no usage uploads except if requested"
                    )
                )
            return

        if self._has_upload_consent(now) and not for_cli_verbose_output:
            self.logger.D("user has consented to telemetry upload")
            return

        if for_cli_verbose_output:
            self.logger.I(
                _(
                    "telemetry mode is [green]on[/]: usage data is collected and periodically uploaded"
                )
            )
            self.logger.I(
                _(
                    "non-tracking usage information will be uploaded to RuyiSDK-managed servers [bold green]every {weekday}[/]"
                ).format(
                    weekday=upload_wday_name,
                )
            )
        else:
            self.logger.W(
                _(
                    "this [yellow]ruyi[/] installation has telemetry mode set to [yellow]on[/], and [bold]will upload non-tracking usage information to RuyiSDK-managed servers[/] [bold green]every {weekday}[/]"
                ).format(
                    weekday=upload_wday_name,
                )
            )

        self._print_upload_schedule_notice(upload_wday, now)

        if not for_cli_verbose_output:
            self.logger.I(_("in order to hide this banner:"))
            self.logger.I(_(" - opt out with [yellow]ruyi telemetry optout[/]"))
            self.logger.I(
                _(" - or give consent with [yellow]ruyi telemetry consent[/]")
            )

    def _next_upload_day(self, time_now: float | None = None) -> int | None:
        upload_wday = self._upload_weekday()
        if upload_wday is None:
            return None
        return next_utc_weekday(upload_wday, time_now)

    def _is_upload_day(self, time_now: float | None = None) -> bool:
        if time_now is None:
            time_now = time.time()
        if upload_day := self._next_upload_day(time_now):
            return upload_day <= time_now
        return False

    def _has_uploaded_today(
        self,
        last_upload_time: float | None,
        time_now: float | None = None,
    ) -> bool:
        if time_now is None:
            time_now = time.time()
        if upload_day := self._next_upload_day(time_now):
            upload_day_end = upload_day + 86400
            if last_upload_time is not None:
                return upload_day <= last_upload_time < upload_day_end
        return False

    def record(self, scope: TelemetryScope, kind: str, **params: object) -> None:
        if self.minimal:
            self.logger.D(
                f"minimal telemetry mode enabled, discarding event '{kind}' for scope {scope}"
            )
            return

        if store := self.store(scope):
            return store.record(kind, **params)
        self.logger.D(
            f"no telemetry store for scope {scope}, discarding event '{kind}'"
        )

    def discard_events(self, v: bool = True) -> None:
        self._discard_events = v

    def _should_proceed_with_upload(
        self,
        scope: TelemetryScope,
        explicit_request: bool,
        cron_mode: bool,
        now: float,
    ) -> tuple[bool, str]:
        # proceed to uploading if forced (explicit requested or _upload_on_exit)
        # regardless of schedule
        if explicit_request:
            return True, "explicit request"
        if self._upload_on_exit:
            return True, "first-run upload on exit"

        # this is not an explicitly requested upload, so only proceed if today
        # is the day, or if the last upload is more than a week ago
        #
        # the last-upload-more-than-a-week-ago check is to avoid situations
        # where the user has not run ruyi for a long time, thus missing
        # the scheduled upload day.
        #
        # cron jobs are a mitigation, but we cannot rely on them either, because:
        #
        # * ruyi is more likely installed user-locally than system-wide, so
        #   users may not set up cron jobs for themselves;
        # * telemetry data is always recorded per user so system-wide cron jobs
        #   cannot easily access this data.
        last_upload_time: float | None = None
        if store := self.store(scope):
            last_upload_time = store.last_upload_timestamp

        if not self._is_upload_day(now):
            if last_upload_time is not None and now - last_upload_time >= 7 * 86400:
                return True, "last upload more than a week ago"
            return False, "not upload day"
        # now we're sure today is the day

        # if we're in cron mode, proceed as if it's an explicit request;
        # otherwise, only proceed if mode is "on" and we haven't uploaded yet today
        # for this scope
        if cron_mode:
            return True, "cron mode upload on upload day"

        if self._gc.telemetry_mode != "on":
            return False, "telemetry mode not 'on'"

        if not self._has_uploaded_today(last_upload_time, now):
            return True, "upload day, not yet uploaded today"
        return False, "upload day, already uploaded today"

    def flush(self, *, upload_now: bool = False, cron_mode: bool = False) -> None:
        """
        Flush collected telemetry data to persistent store, and upload if needed.

        :param upload_now: Upload data right now regardless of schedule.
        :type upload_now: bool
        :param cron_mode: Whether this flush is called from a cron job. If true,
            non-upload-day uploads will be skipped, otherwise acts just like
            explicit uploads via `ruyi telemetry upload`.
        :type cron_mode: bool
        """

        # We may be self-uninstalling and purging all state data, and in this
        # case we don't want to record anything (thus re-creating directories).
        if self._discard_events:
            self.logger.D("discarding collected telemetry data")
            return

        now = time.time()

        def should_proceed(scope: TelemetryScope) -> tuple[bool, str]:
            return self._should_proceed_with_upload(
                scope,
                explicit_request=upload_now,
                cron_mode=cron_mode,
                now=now,
            )

        if self.minimal:
            if not self._upload_on_exit:
                self.logger.D("skipping upload for non-first-run in minimal mode")
                return

            for scope, store in self._stores.items():
                go_ahead, reason = should_proceed(scope)
                self.logger.D(
                    f"minimal telemetry upload check for scope {scope}: go_ahead={go_ahead}, reason={reason}"
                )
                if not go_ahead:
                    continue
                store.upload_minimal()
            return

        for scope, store in self._stores.items():
            self.logger.D(f"flushing telemetry to persistent store for scope {scope}")
            store.persist(now)

            go_ahead, reason = should_proceed(scope)
            self.logger.D(
                f"regular telemetry upload check for scope {scope}: go_ahead={go_ahead}, reason={reason}"
            )
            if not go_ahead:
                continue

            self._prepare_data_for_upload(store)
            store.upload_staged_payloads()

    def _prepare_data_for_upload(self, store: TelemetryStore) -> None:
        installation_data: NodeInfo | None = None
        if store.scope.is_pm:
            try:
                installation_data = self._read_installation_data()
            except FileNotFoundError:
                # should not happen due to is_upload_day() initializing it for us
                # beforehand, but proceed without node info nonetheless
                pass

        return store.prepare_data_for_upload(installation_data)

    def oobe_prompt(self) -> None:
        """Ask whether the user consents to a first-run telemetry upload, and
        persist the user's exact telemetry choice."""

        if self._gc.is_telemetry_optout:
            # user has already explicitly opted out via the environment variable,
            # don't bother asking
            return

        # We always report installation info on first run, regardless of
        # user's telemetry choice. In case the user opts out, only do a one-time
        # upload now, and never upload anything again.
        self._upload_on_exit = True

        from ..cli import user_input

        self.logger.stdout(_(TELEMETRY_CONSENT_AND_UPLOAD_DESC))
        if not user_input.ask_for_yesno_confirmation(
            self.logger,
            _(TELEMETRY_CONSENT_AND_UPLOAD_PROMPT),
            False,
        ):
            # ask if the user wants to opt out entirely
            if user_input.ask_for_yesno_confirmation(
                self.logger,
                _(TELEMETRY_OPTOUT_PROMPT),
                False,
            ):
                set_telemetry_mode(self._gc, "off")
                return

            # user wants to stay in local mode
            # explicitly record the preference, so we don't have to worry about
            # us potentially changing defaults yet another time
            set_telemetry_mode(self._gc, "local")
            return

        consent_time = datetime.datetime.now().astimezone()
        set_telemetry_mode(self._gc, "on", consent_time)



import argparse
import datetime
from typing import TYPE_CHECKING

from ..cli.cmd import RootCommand
from ..i18n import _

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


# Telemetry preference commands
class TelemetryCommand(
    RootCommand,
    cmd="telemetry",
    has_main=True,
    has_subcommands=True,
    help=_("Manage your telemetry preferences"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        # https://github.com/python/cpython/issues/67037 prevents the registration
        # of undocumented subcommands, so a preferred usage of
        # "ruyi telemetry cron-upload" is not possible right now.
        p.add_argument(
            "--cron-upload",
            action="store_true",
            dest="cron_upload",
            default=False,
            help=argparse.SUPPRESS,
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        cron_upload: bool = args.cron_upload
        if not cron_upload:
            args._parser.print_help()  # pylint: disable=protected-access
            return 0

        # the rest are implementation of "--cron-upload"

        cfg.telemetry.flush(cron_mode=True)
        return 0


class TelemetryConsentCommand(
    TelemetryCommand,
    cmd="consent",
    aliases=["on"],
    help=_("Give consent to telemetry data uploads"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .provider import set_telemetry_mode

        now = datetime.datetime.now().astimezone()
        set_telemetry_mode(cfg, "on", now)
        return 0


class TelemetryLocalCommand(
    TelemetryCommand,
    cmd="local",
    help=_("Set telemetry mode to local collection only"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .provider import set_telemetry_mode

        set_telemetry_mode(cfg, "local")
        return 0


class TelemetryOptoutCommand(
    TelemetryCommand,
    cmd="optout",
    aliases=["off"],
    help=_("Opt out of telemetry data collection"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .provider import set_telemetry_mode

        set_telemetry_mode(cfg, "off")
        return 0


class TelemetryStatusCommand(
    TelemetryCommand,
    cmd="status",
    help=_("Print the current telemetry mode"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument(
            "--verbose",
            "-v",
            action="store_true",
            help=_("Enable verbose output"),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        verbose: bool = args.verbose
        if not verbose:
            cfg.logger.stdout(cfg.telemetry_mode)
            return 0

        if cfg.telemetry is None:
            cfg.logger.I(
                _("telemetry mode is [green]off[/]: no further data will be collected")
            )
            return 0

        cfg.telemetry.print_telemetry_notice(for_cli_verbose_output=True)
        return 0


class TelemetryUploadCommand(
    TelemetryCommand,
    cmd="upload",
    help=_("Upload collected telemetry data now"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        cfg.telemetry.flush(upload_now=True)
        # disable the flush at program exit because we have just done that
        cfg.telemetry.discard_events()
        return 0



#!/usr/bin/env python3

import os
import platform
import shutil
import subprocess
import sys
import tomllib
from typing import cast

PYGIT2_SETUPTOOLS_PATCH = """
--- a/build.sh
+++ b/build.sh
@@ -82,6 +82,8 @@ if [ "$CIBUILDWHEEL" = "1" ]; then
 else
     # Create a virtual environment
     $PYTHON -m venv $PREFIX
+    # install setuptools in the venv for python3.12+
+    $PREFIX/bin/pip install -U setuptools pycparser
     cd ci
 fi
 
"""

PYGIT2_OPENSSL_NO_DOCS_PATCH = """
--- a/build.sh
+++ b/build.sh
@@ -134,7 +134,7 @@ if [ -n "$OPENSSL_VERSION" ]; then
         # Linux
         tar xf $FILENAME.tar.gz
         cd $FILENAME
-        ./Configure shared --prefix=$PREFIX --libdir=$PREFIX/lib
+        ./Configure shared no-apps no-docs no-tests --prefix=$PREFIX --libdir=$PREFIX/lib
         make
         make install
         OPENSSL_PREFIX=$(pwd)
"""


def get_pygit2_src_uri(tag: str) -> tuple[str, str]:
    filename = f"{tag}.tar.gz"
    return (filename, f"https://github.com/libgit2/pygit2/archive/refs/tags/{filename}")


def log(s: str, fgcolor: int = 32, group: bool = False) -> None:
    # we cannot import rich because this script is executed before
    # `poetry install` in the dist build process
    print(f"\x1b[1;{fgcolor}m{s}\x1b[m", file=sys.stderr, flush=True)
    if group:
        begin_group(s)


def is_in_gha() -> bool:
    return "GITHUB_ACTIONS" in os.environ


def begin_group(title: str) -> None:
    if is_in_gha():
        print(f"::group::{title}", flush=True)


def end_group() -> None:
    if is_in_gha():
        print("::endgroup::", flush=True)


def main() -> None:
    build_root = os.environ["RUYI_DIST_BUILD_DIR"]

    workdir = os.path.join(build_root, "ruyi-pygit2")
    ensure_dir(workdir)

    cache_root = os.environ["RUYI_DIST_CACHE_DIR"]
    ensure_dir(cache_root)

    pygit2_ver = get_pygit2_version()
    log(f"resolved pygit2 version {pygit2_ver}")

    pygit2_cache_rev = 1  # bump this to force rebuild (e.g. for bumping indirect deps)
    pygit2_wheel_path = ensure_pygit2_wheel(
        pygit2_ver,
        workdir,
        cache_root,
        pygit2_cache_rev,
    )

    # this will print a header suitable for our logging purposes
    begin_group("pip install the pygit2 wheel")
    subprocess.run(("pip", "install", pygit2_wheel_path), check=True)
    end_group()

    log("informing poetry about the wheel", group=True)
    subprocess.run(("poetry", "add", "--lock", pygit2_wheel_path), check=True)
    end_group()


def ensure_dir(d: str) -> None:
    try:
        os.mkdir(d)
    except FileExistsError:
        pass


def ensure_pygit2_wheel(ver: str, workdir: str, cache_root: str, cache_rev: int) -> str:
    cache_key = get_cache_key("pygit2", ver, cache_rev)
    cache_dir = os.path.join(cache_root, cache_key)
    try:
        wheel_file = [
            i for i in os.listdir(cache_dir) if os.path.splitext(i)[1] == ".whl"
        ][0]
        wheel_path = os.path.join(cache_dir, wheel_file)
        log(f"found cached pygit2 (cache rev {cache_rev}) at {wheel_path}")
    except (FileNotFoundError, IndexError):
        log(f"cached pygit2 (cache rev {cache_rev}) not found, building")
        wheel_path = build_pygit2(ver, workdir)

        log(f"caching built wheel {wheel_path} to {cache_dir}")
        ensure_dir(cache_dir)

        dest_path = os.path.join(cache_dir, os.path.basename(wheel_path))
        shutil.copyfile(wheel_path, dest_path)

    return wheel_path


def build_pygit2(pygit2_ver: str, workdir: str) -> str:
    pygit2_tag = f"v{pygit2_ver}"
    pygit2_src_filename, pygit2_src_uri = get_pygit2_src_uri(pygit2_tag)
    pygit2_workdir = os.path.join(workdir, f"pygit2-{pygit2_ver}")

    # download the source
    log(f"downloading {pygit2_src_uri}", group=True)
    subprocess.run(
        ("wget", "-O", pygit2_src_filename, pygit2_src_uri),
        cwd=workdir,
        check=True,
    )
    end_group()

    # unpack the source
    log(f"unpacking {pygit2_src_filename}")
    subprocess.run(("tar", "-xf", pygit2_src_filename), cwd=workdir, check=True)

    log("patching pygit2 for Python 3.12+ builds")
    subprocess.run(
        ("patch", "-Np1"),
        cwd=pygit2_workdir,
        input=PYGIT2_SETUPTOOLS_PATCH.encode("utf-8"),
        check=True,
    )

    log("disabling docs generation during pygit2 openssl build")
    subprocess.run(
        ("patch", "-Np1"),
        cwd=pygit2_workdir,
        input=PYGIT2_OPENSSL_NO_DOCS_PATCH.encode("utf-8"),
        check=True,
    )

    # build wheel
    extra_env = get_pygit2_wheel_build_env(pygit2_workdir)
    log("extra envvar(s):", fgcolor=36)
    for k, v in extra_env.items():
        log(f"  {k}: {v}", fgcolor=36)
        os.environ[k] = v

    log("building pygit2 wheel", group=True)
    subprocess.run(
        ("sh", "build.sh", "wheel", "bundle"),
        cwd=pygit2_workdir,
        check=True,
    )
    end_group()

    pygit2_distdir = os.path.join(pygit2_workdir, "wheelhouse")
    pygit2_wheel_name = find_built_wheel_name_in(pygit2_distdir)
    return os.path.join(pygit2_distdir, pygit2_wheel_name)


def get_pygit2_version() -> str:
    # assume CWD is project root, which is guaranteed to be the case (see
    # end of file)
    with open("poetry.lock", "rb") as fp:
        info = tomllib.load(fp)

    pygit2 = [pkg for pkg in info["package"] if pkg["name"] == "pygit2"][0]
    return cast(str, pygit2["version"])


def get_pygit2_wheel_build_env(pygit2_dir: str) -> dict[str, str]:
    with open(os.path.join(pygit2_dir, "pyproject.toml"), "rb") as fp:
        pyproject = tomllib.load(fp)

    r: dict[str, str] = pyproject["tool"]["cibuildwheel"]["environment"]
    if "LIBGIT2" in r:
        # this is unnecessary
        del r["LIBGIT2"]

    if platform.machine() == "riscv64":
        # auditwheel 6.1.0+ has manylinux policies for riscv64, but the default
        # is too low for our environment
        # bump it up
        r["AUDITWHEEL_PLAT"] = "manylinux_2_35_riscv64"

    # This controls the build type for the CMake-built libgit2, and defaults to
    # Debug otherwise in the pygit2 build.sh.
    r["BUILD_TYPE"] = "Release"

    return r


def find_built_wheel_name_in(path: str) -> str:
    return [x for x in os.listdir(path) if os.path.splitext(x)[1] == ".whl"][0]


def get_cache_key(pkg_name: str, version: str, cache_rev: int) -> str:
    return f"{pkg_name}-{version}-cache{cache_rev}"


if __name__ == "__main__":
    # cd to project root
    os.chdir(os.path.join(os.path.dirname(__file__), ".."))
    main()



import os.path

from ..i18n import _
from ..log import RuyiLogger


def pause_before_continuing(
    logger: RuyiLogger,
) -> None:
    """Pause and wait for the user to press Enter before continuing.

    EOFError should be handled by the caller."""

    logger.stdout(_("Press [green]<ENTER>[/] to continue: "), end="")
    input()


def ask_for_yesno_confirmation(
    logger: RuyiLogger,
    prompt: str,
    default: bool = False,
) -> bool:
    choices_help = "(Y/n)" if default else "(y/N)"

    while True:
        try:
            logger.stdout(f"{prompt} {choices_help} ", end="")
            user_input = input()
        except EOFError:
            yesno = _("YES") if default else _("NO")
            logger.W(
                _(
                    "EOF while reading user input, assuming the default choice {yesno}"
                ).format(yesno=yesno)
            )
            return default

        if not user_input:
            return default
        if user_input in {"Y", "y", "yes"}:
            return True
        if user_input in {"N", "n", "no"}:
            return False
        else:
            logger.stdout(
                _("Unrecognized input [yellow]'{user_input}'[/].").format(
                    user_input=user_input
                )
            )
            logger.stdout(_("Accepted choices: Y/y/yes for YES, N/n/no for NO."))


def ask_for_kv_choice(
    logger: RuyiLogger,
    prompt: str,
    choices_kv: dict[str, str],
    default_key: str | None = None,
) -> str:
    choices_kv_list = list(choices_kv.items())
    choices_prompts = [i[1] for i in choices_kv_list]

    default_idx: int | None = None
    if default_key is not None:
        for i, k in enumerate(choices_kv_list):
            if k[0] == default_key:
                default_idx = i
                break
        if default_idx is None:
            raise ValueError(f"Default choice key '{default_key}' not in choices")

    choice = ask_for_choice(logger, prompt, choices_prompts, default_idx)
    return choices_kv_list[choice][0]


def ask_for_choice(
    logger: RuyiLogger,
    prompt: str,
    choices_texts: list[str],
    default_idx: int | None = None,
) -> int:
    logger.stdout(prompt, end="\n\n")
    for i, choice_text in enumerate(choices_texts):
        logger.stdout(f"  {i + 1}. {choice_text}")

    logger.stdout("")

    nr_choices = len(choices_texts)
    if default_idx is not None:
        if not (0 <= default_idx < nr_choices):
            raise ValueError(f"Default choice index {default_idx} out of range")
        choices_help = _("(1-{nr_choices}, default {default})").format(
            nr_choices=nr_choices,
            default=default_idx + 1,
        )
    else:
        choices_help = _("(1-{nr_choices})").format(nr_choices=nr_choices)
    while True:
        try:
            user_input = input(
                _("Choice? {choices_help} ").format(
                    choices_help=choices_help,
                )
            )
        except EOFError:
            raise ValueError("EOF while reading user choice")

        if default_idx is not None and not user_input:
            return default_idx

        try:
            choice_int = int(user_input)
        except ValueError:
            logger.stdout(
                _("Unrecognized input [yellow]'{user_input}'[/].").format(
                    user_input=user_input,
                )
            )
            logger.stdout(
                _(
                    "Accepted choices: an integer number from 1 to {nr_choices} inclusive."
                ).format(
                    nr_choices=nr_choices,
                )
            )
            continue

        if 1 <= choice_int <= nr_choices:
            return choice_int - 1

        logger.stdout(
            _("Out-of-range input [yellow]'{user_input}'[/].").format(
                user_input=user_input,
            )
        )
        logger.stdout(
            _(
                "Accepted choices: an integer number from 1 to {nr_choices} inclusive."
            ).format(
                nr_choices=nr_choices,
            )
        )


def ask_for_file(
    logger: RuyiLogger,
    prompt: str,
) -> str:
    while True:
        try:
            user_input = input(f"{prompt} ")
        except EOFError:
            raise ValueError("EOF while reading user input")

        if os.path.exists(user_input):
            return user_input

        logger.stdout(f"[yellow]'{user_input}'[/] is not a path to an existing file.")



from functools import cached_property
import json
import os
import pathlib
import re
import time
from typing import Callable, Final, Iterable
import uuid

from ..log import RuyiLogger
from ..utils.node_info import NodeInfo
from ..utils.url import urljoin_for_sure
from .aggregate import UploadPayload, aggregate_events
from .event import TelemetryEvent, is_telemetry_event
from .scope import TelemetryScope

# e.g. "run.202410201845.d06ca5d668e64fec833ed3e6eb926a2c.ndjson"
RE_RAW_EVENT_FILENAME: Final = re.compile(
    r"^run\.(?P<time_bucket>\d{12})\.(?P<uuid>[0-9a-f]{32})\.ndjson$"
)


def get_time_bucket(timestamp: int | float | time.struct_time | None = None) -> str:
    if timestamp is None:
        return time.strftime("%Y%m%d%H%M")
    elif isinstance(timestamp, float) or isinstance(timestamp, int):
        timestamp = time.localtime(timestamp)
    return time.strftime("%Y%m%d%H%M", timestamp)


def time_bucket_from_filename(filename: str) -> str | None:
    if m := RE_RAW_EVENT_FILENAME.match(filename):
        return m.group("time_bucket")
    return None


class TelemetryStore:
    def __init__(
        self,
        logger: RuyiLogger,
        scope: TelemetryScope,
        store_root: pathlib.Path,
        api_url: str | None = None,
        api_url_factory: Callable[[], str | None] | None = None,
    ) -> None:
        self._logger = logger
        self.scope = scope
        self.store_root = store_root
        self._api_url = api_url
        self._api_url_factory = api_url_factory

        self._events: list[TelemetryEvent] = []

    @cached_property
    def api_url(self) -> str | None:
        if u := self._api_url:
            return u
        if f := self._api_url_factory:
            return f()
        return None

    @property
    def raw_events_dir(self) -> pathlib.Path:
        return self.store_root / "raw"

    @property
    def upload_stage_dir(self) -> pathlib.Path:
        return self.store_root / "staged"

    @property
    def uploaded_dir(self) -> pathlib.Path:
        return self.store_root / "uploaded"

    @property
    def last_upload_marker_file(self) -> pathlib.Path:
        return self.store_root / ".stamp-last-upload"

    @property
    def last_upload_timestamp(self) -> float | None:
        try:
            return self.last_upload_marker_file.stat().st_mtime
        except FileNotFoundError:
            return None

    def record_upload_timestamp(self, time_now: float | None = None) -> None:
        if time_now is None:
            time_now = time.time()

        # We may not have store_root existing yet if we're in minimal telemetry
        # mode
        self.store_root.mkdir(parents=True, exist_ok=True)

        f = self.last_upload_marker_file
        f.touch()
        os.utime(f, (time_now, time_now))

    def record(self, kind: str, **params: object) -> None:
        self._events.append({"fmt": 1, "kind": kind, "params": params})

    def discard_events(self, v: bool = True) -> None:
        self._discard_events = v

    def persist(self, now: float | None = None) -> None:
        if not self._events:
            self._logger.D(f"scope {self.scope}: no event to persist")
            return

        now = time.time() if now is None else now

        self._logger.D(f"scope {self.scope}: flushing telemetry to persistent store")

        raw_events_dir = self.raw_events_dir
        raw_events_dir.mkdir(parents=True, exist_ok=True)

        # TODO: for now it is safe to not lock, because flush() is only ever
        # called at program exit time
        rough_time = get_time_bucket(now)
        rand = uuid.uuid4().hex
        batch_events_file = raw_events_dir / f"run.{rough_time}.{rand}.ndjson"
        with open(batch_events_file, "wb") as fp:
            for e in self._events:
                payload = json.dumps(e)
                fp.write(payload.encode("utf-8"))
                fp.write(b"\n")

        self._logger.D(
            f"scope {self.scope}: persisted {len(self._events)} telemetry event(s)"
        )

    def read_back_raw_events(self) -> Iterable[TelemetryEvent]:
        try:
            for f in self.raw_events_dir.glob("run.*.ndjson"):
                time_bucket = time_bucket_from_filename(f.name)
                with open(f, "r", encoding="utf-8", newline=None) as fp:
                    for line in fp:
                        try:
                            obj = json.loads(line)
                        except json.JSONDecodeError:
                            # losing some malformed telemetry events is okay
                            continue
                        if not is_telemetry_event(obj):
                            # ditto
                            continue
                        if time_bucket is not None and "time_bucket" not in obj:
                            obj["time_bucket"] = time_bucket
                        yield obj
        except FileNotFoundError:
            pass

    def purge_raw_events(self) -> None:
        files = list(self.raw_events_dir.glob("run.*.ndjson"))
        for f in files:
            f.unlink(missing_ok=True)

    def gen_upload_staging_filename(self, nonce: str) -> pathlib.Path:
        return self.upload_stage_dir / f"staged.{nonce}.json"

    def prepare_data_for_upload(self, installation_data: NodeInfo | None) -> None:
        # import ruyi.version here because this package is on the CLI startup
        # critical path, and version probing is costly there
        from ..version import RUYI_SEMVER

        aggregate_data = list(aggregate_events(self.read_back_raw_events()))

        payload_nonce = uuid.uuid4().hex  # for server-side dedup purposes
        payload: UploadPayload = {
            "fmt": 1,
            "nonce": payload_nonce,
            "ruyi_version": RUYI_SEMVER,
            "events": aggregate_data,
        }
        if installation_data is not None:
            payload["installation"] = installation_data

        dest_path = self.gen_upload_staging_filename(payload_nonce)
        self.upload_stage_dir.mkdir(parents=True, exist_ok=True)
        dest_path.write_text(json.dumps(payload), encoding="utf-8")

        self.purge_raw_events()

    def prepare_data_for_minimal_upload(self) -> bytes:
        """Prepare a minimal upload payload with no installation data and no events.

        Used when user has not consented to telemetry collection but also not
        explicitly opted out, for gaining minimal insight into adoption.
        """

        # import ruyi.version here because this package is on the CLI startup
        # critical path, and version probing is costly there
        from ..version import RUYI_SEMVER

        payload_nonce = uuid.uuid4().hex  # for server-side dedup purposes

        # We don't have installation data, and cannot have it initialized
        # in this case because absence of installation data means no user
        # consent. And making up a persistent installation ID is not a
        # choice either, because "installation ID" resembles "advertising
        # ID" a lot, which is considered personally identifiable information
        # (PII) and not allowed to be collected without user consent.
        #
        # So, resort to re-using the completely random nonce as the report
        # UUID, which does not allow for server-side correlation but at least
        # allows for some insight into end-user adoption.
        payload: UploadPayload = {
            "fmt": 1,
            "nonce": payload_nonce,
            "ruyi_version": RUYI_SEMVER,
            "report_uuid": payload_nonce,
            "events": [],
        }

        return json.dumps(payload).encode("utf-8")

    def upload_minimal(self) -> None:
        if not self.api_url:
            return

        p = self.prepare_data_for_minimal_upload()
        self.upload_one_staged_payload(p, self.api_url)
        self.record_upload_timestamp()

    def upload_staged_payloads(self) -> None:
        if not self.api_url:
            return

        try:
            staged_payloads = list(self.upload_stage_dir.glob("staged.*.json"))
        except FileNotFoundError:
            return

        try:
            self.uploaded_dir.mkdir(parents=True, exist_ok=True)
        except OSError:
            return

        for f in staged_payloads:
            self.upload_one_staged_payload(f, self.api_url)

        self.record_upload_timestamp()

    def upload_one_staged_payload(
        self,
        f: pathlib.Path | bytes,
        endpoint: str,
    ) -> None:
        # import ruyi.version here because this package is on the CLI startup
        # critical path, and version probing is costly there
        from ..version import RUYI_USER_AGENT

        api_path = urljoin_for_sure(endpoint, "upload-v1")

        if isinstance(f, pathlib.Path):
            self._logger.D(
                f"scope {self.scope}: about to upload payload {f} to {api_path}"
            )
            data = f.read_bytes()
        else:
            self._logger.D(
                f"scope {self.scope}: about to upload in-memory payload to {api_path}"
            )
            data = f

        import requests

        resp = requests.post(
            api_path,
            data=data,
            headers={"User-Agent": RUYI_USER_AGENT},
            allow_redirects=True,
            timeout=5,
        )

        if not (200 <= resp.status_code < 300):
            self._logger.D(
                f"scope {self.scope}: telemetry upload failed: status code {resp.status_code}, content {resp.content.decode('utf-8', 'replace')}"
            )
            return

        self._logger.D(
            f"scope {self.scope}: telemetry upload ok: status code {resp.status_code}"
        )

        if isinstance(f, pathlib.Path):
            # move to completed dir
            # TODO: rotation
            try:
                f.rename(self.uploaded_dir / f.name)
            except OSError as e:
                self._logger.D(
                    f"scope {self.scope}: failed to move uploaded payload away: {e}"
                )



import tomlkit

from ruyi.utils.toml import (
    extract_footer_comments,
    extract_header_comments,
)


def test_extract_header_comments() -> None:
    a1 = """[foo]
    a = 2
    """
    assert extract_header_comments(tomlkit.parse(a1)) == []

    a2 = """

# foo

# bar

a = 2
# baz
"""

    assert extract_header_comments(tomlkit.parse(a2)) == [
        "# foo\n",
        "\n",
        "# bar\n",
        "\n",
    ]

    a3 = """
# baz
# quux

[a]
b = 2
"""

    assert extract_header_comments(tomlkit.parse(a3)) == ["# baz\n", "# quux\n", "\n"]


def test_extract_footer_comments() -> None:
    a1 = """[foo]
    a = 2
    """
    assert extract_footer_comments(tomlkit.parse(a1)) == []

    a2 = """
a = 2
# foo
# bar
"""

    assert extract_footer_comments(tomlkit.parse(a2)) == ["# foo\n", "# bar\n"]

    a3 = """# foo

# bar

[a]
foo = 2

# baz
# baz2

# quux
"""

    assert extract_footer_comments(tomlkit.parse(a3)) == [
        "\n",
        "# baz\n",
        "# baz2\n",
        "\n",
        "# quux\n",
    ]



import os
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self

from ...ruyipkg.pkg_manifest import EmulatorProgDecl
from ...ruyipkg.profile import ProfileProxy


class ResolvedEmulatorProg:
    def __init__(
        self,
        display_name: str,
        binfmt_misc_str: str | None,
        env: dict[str, str] | None,
    ) -> None:
        self.display_name = display_name
        self.binfmt_misc_str = binfmt_misc_str
        self.env = env

    @classmethod
    def new(
        cls,
        prog: EmulatorProgDecl,
        prog_install_root: os.PathLike[Any],
        profile: ProfileProxy,
        sysroot: os.PathLike[Any] | None,
    ) -> "Self":
        return cls(
            get_display_name_for_emulator(prog, prog_install_root),
            prog.get_binfmt_misc_str(prog_install_root),
            profile.get_env_config_for_emu_flavor(prog.flavor, sysroot),
        )


def get_display_name_for_emulator(
    prog: EmulatorProgDecl,
    prog_install_root: os.PathLike[Any],
) -> str:
    return f"{os.path.basename(prog.relative_path)} from {prog_install_root}"
